import itertools
from typing import (
    Any,
    Dict,
)

from eth_typing import (
    HexStr,
    TypeStr,
)
from eth_utils import (
    function_abi_to_4byte_selector,
    is_0x_prefixed,
    is_binary_address,
    is_boolean,
    is_bytes,
    is_checksum_address,
    is_dict,
    is_hex_address,
    is_integer,
    is_list_like,
    is_string,
)
from eth_utils.curried import (
    apply_formatter_to_array,
)
from eth_utils.hexadecimal import (
    encode_hex,
)
from eth_utils.toolz import (
    compose,
    groupby,
    valfilter,
    valmap,
)

from cns.utils import (
    is_valid_cns_name,
)
from web3._utils.abi import (
    abi_to_signature,
    filter_by_type,
    is_address_type,
    is_array_type,
    is_bool_type,
    is_bytes_type,
    is_int_type,
    is_recognized_type,
    is_string_type,
    is_uint_type,
    length_of_array_type,
    sub_type_of_array_type,
)
from web3.exceptions import (
    InvalidAddress,
)
from web3.types import (  # noqa: F401
    ABI,
    ABIEvent,
    ABIFunction,
)


def _prepare_selector_collision_msg(duplicates: Dict[HexStr, ABIFunction]) -> str:
    dup_sel = valmap(apply_formatter_to_array(abi_to_signature), duplicates)
    joined_funcs = valmap(lambda funcs: ", ".join(funcs), dup_sel)
    func_sel_msg_list = [
        funcs + " have selector " + sel for sel, funcs in joined_funcs.items()
    ]
    return " and\n".join(func_sel_msg_list)


def validate_abi(abi: ABI) -> None:
    """
    Helper function for validating an ABI
    """
    if not is_list_like(abi):
        raise ValueError("'abi' is not a list")

    if not all(is_dict(e) for e in abi):
        raise ValueError("'abi' is not a list of dictionaries")

    functions = filter_by_type("function", abi)
    selectors = groupby(compose(encode_hex, function_abi_to_4byte_selector), functions)
    duplicates = valfilter(lambda funcs: len(funcs) > 1, selectors)
    if duplicates:
        raise ValueError(
            "Abi contains functions with colliding selectors. "
            f"Functions {_prepare_selector_collision_msg(duplicates)}"
        )


def validate_abi_type(abi_type: TypeStr) -> None:
    """
    Helper function for validating an abi_type
    """
    if not is_recognized_type(abi_type):
        raise ValueError(f"Unrecognized abi_type: {abi_type}")


def validate_abi_value(abi_type: TypeStr, value: Any) -> None:
    """
    Helper function for validating a value against the expected abi_type
    Note: abi_type 'bytes' must either be python3 'bytes' object or ''
    """
    if is_array_type(abi_type) and is_list_like(value):
        # validate length
        specified_length = length_of_array_type(abi_type)
        if specified_length is not None:
            if specified_length < 1:
                raise TypeError(
                    f"Invalid abi-type: {abi_type}. Length of fixed sized "
                    "arrays must be greater than 0."
                )
            if specified_length != len(value):
                raise TypeError(
                    "The following array length does not the length specified"
                    f"by the abi-type, {abi_type}: {value}"
                )

        # validate sub_types
        sub_type = sub_type_of_array_type(abi_type)
        for v in value:
            validate_abi_value(sub_type, v)
        return
    elif is_bool_type(abi_type) and is_boolean(value):
        return
    elif is_uint_type(abi_type) and is_integer(value) and value >= 0:
        return
    elif is_int_type(abi_type) and is_integer(value):
        return
    elif is_address_type(abi_type):
        validate_address(value)
        return
    elif is_bytes_type(abi_type):
        if is_bytes(value):
            return
        elif is_string(value):
            if is_0x_prefixed(value):
                return
            else:
                raise TypeError(
                    "ABI values of abi-type 'bytes' must be either"
                    "a python3 'bytes' object or an '0x' prefixed string."
                )
    elif is_string_type(abi_type) and is_string(value):
        return

    raise TypeError(f"The following abi value is not a '{abi_type}': {value}")


def is_not_address_string(value: Any) -> bool:
    return (
        is_string(value)
        and not is_bytes(value)
        and not is_checksum_address(value)
        and not is_hex_address(value)
    )


def validate_address(value: Any) -> None:
    """
    Helper function for validating an address
    """
    if is_not_address_string(value):
        if not is_valid_cns_name(value):
            raise InvalidAddress(f"CNS name: '{value}' is invalid.")
        return
    if is_bytes(value):
        if not is_binary_address(value):
            raise InvalidAddress(
                "Address must be 20 bytes when input type is bytes", value
            )
        return

    if not isinstance(value, str):
        raise TypeError(f"Address {value} must be provided as a string")
    if not is_hex_address(value):
        raise InvalidAddress(
            "Address must be 20 bytes, as a hex string with a 0x prefix", value
        )
    if not is_checksum_address(value):
        if value == value.lower():
            raise InvalidAddress(
                "Web3.py only accepts checksum addresses. "
                "The software that gave you this non-checksum address should be "
                "considered unsafe, please file it as a bug on their platform. "
                "Try using an ENS name instead. Or, if you must accept lower safety, "
                "use Web3.toChecksumAddress(lower_case_address).",
                value,
            )
        else:
            raise InvalidAddress(
                "Address has an invalid EIP-55 checksum. "
                "After looking up the address from the original source, try again.",
                value,
            )


def has_one_val(*args: Any, **kwargs: Any) -> bool:
    vals = itertools.chain(args, kwargs.values())
    not_nones = list(filter(lambda val: val is not None, vals))
    return len(not_nones) == 1


def assert_one_val(*args: Any, **kwargs: Any) -> None:
    if not has_one_val(*args, **kwargs):
        raise TypeError(
            "Exactly one of the passed values can be specified. "
            f"Instead, values were: {args!r}, {kwargs!r}"
        )



import re
from typing import (
    Any,
    Dict,
)

from eth_typing import (
    Manifest,
)
from eth_utils.toolz import (
    assoc,
    assoc_in,
    curry,
)

from ethpm.constants import (
    PACKAGE_NAME_REGEX,
)
from ethpm.tools.builder import (
    build,
)

# todo: validate no duplicate blockchain uris in deployments, if web3 is available

WARNINGS = {
    "manifest_missing": "Manifest missing a required 'manifest' field.",
    "manifest_invalid": """'manifest' is invalid. The only supported"""
    """ version is 'ethpm/3'.""",
    "name_missing": "Manifest missing a suggested 'name' field",
    "name_invalid": "'name' is invalid. "
    f"Doesn't match the regex: {PACKAGE_NAME_REGEX}",
    "version_missing": "Manifest missing a suggested 'version' field.",
    "meta_missing": "Manifest missing a suggested 'meta' field.",
    "authors_missing": "'meta' field missing suggested 'authors' field.",
    "description_missing": "'meta' field missing suggested 'description' field.",
    "links_missing": "'meta' field missing suggested 'links' field.",
    "license_missing": "'meta' field missing suggested 'license' field.",
    "keywords_missing": "'meta' field missing suggested 'keywords' field.",
    "sources_missing": """Manifest is missing a sources field, """
    """which defines a source tree that should comprise the full source tree """
    """necessary to recompile the contracts contained in this release.""",
    "contract_type_missing": """Manifest does not contain any 'contractTypes'. """
    """Packages should only include contract types that can be found in the """
    """source files for this package. Packages should not include contract types """
    """from dependencies. Packages should not include abstract contracts in the """
    """contract types section of a release.""",
    "abi_missing": """Contract type: {0} is missing an abi field, which is """
    """essential for using this package.""",
    "deployment_bytecode_missing": """Contract type: {0} is missing a """
    """`deploymentBytecode` field, which is essential for using this package.""",
    "contract_type_subfield_missing": """Contract type: {0} is missing a"""
    """ `contractType` field, which is essential if an alias is being used """
    """to namespace this contract type.""",
    "runtime_bytecode_missing": """Contract type: {0} is missing a """
    """`runtimeBytecode` field.""",
    "bytecode_subfield_missing": """Contract type: {0} is missing a required  """
    """bytecode subfield in its {1} bytecode object.""",
    "devdoc_missing": "Contract type: {0} is missing a devdoc field.",
    "userdoc_missing": "Contract type: {0} is missing a userdoc field.",
    "compilers_missing": "Manifest is missing a suggested `compilers` field.",
}


#
# Validation
#


def check_manifest(manifest: Manifest) -> Dict[str, str]:
    generate_warnings = (
        check_manifest_version(manifest),
        check_package_name(manifest),
        check_version(manifest),
        check_meta(manifest),
        check_sources(manifest),
        check_contract_types(manifest),
        check_compilers(manifest),
    )
    return build({}, *generate_warnings)


#
# Required fields
#


@curry
def check_manifest_version(
    manifest: Manifest, warnings: Dict[str, str]
) -> Dict[str, str]:
    if "manifest" not in manifest or not manifest["manifest"]:
        return assoc(warnings, "manifest", WARNINGS["manifest_missing"])
    if manifest["manifest"] != "ethpm/3":
        return assoc(warnings, "manifest", WARNINGS["manifest_invalid"])
    return warnings


@curry
def check_package_name(manifest: Manifest, warnings: Dict[str, str]) -> Dict[str, str]:
    if "name" not in manifest or not manifest["name"]:
        return assoc(warnings, "name", WARNINGS["name_missing"])
    if not bool(re.match(PACKAGE_NAME_REGEX, manifest["name"])):
        return assoc(warnings, "name", WARNINGS["name_invalid"])
    return warnings


@curry
def check_version(manifest: Manifest, warnings: Dict[str, str]) -> Dict[str, str]:
    if "version" not in manifest or not manifest["version"]:
        return assoc(warnings, "version", WARNINGS["version_missing"])
    return warnings


#
# Meta fields
#


@curry
def check_meta(manifest: Manifest, warnings: Dict[str, str]) -> Dict[str, str]:
    if "meta" not in manifest or not manifest["meta"]:
        return assoc(warnings, "meta", WARNINGS["meta_missing"])
    meta_validation = (
        check_authors(manifest["meta"]),
        check_license(manifest["meta"]),
        check_description(manifest["meta"]),
        check_keywords(manifest["meta"]),
        check_links(manifest["meta"]),
    )
    return build(warnings, *meta_validation)


@curry
def check_authors(meta: Dict[str, Any], warnings: Dict[str, str]) -> Dict[str, str]:
    if "authors" not in meta:
        return assoc(warnings, "meta.authors", WARNINGS["authors_missing"])
    return warnings


@curry
def check_license(meta: Dict[str, Any], warnings: Dict[str, str]) -> Dict[str, str]:
    if "license" not in meta or not meta["license"]:
        return assoc(warnings, "meta.license", WARNINGS["license_missing"])
    return warnings


@curry
def check_description(meta: Dict[str, Any], warnings: Dict[str, str]) -> Dict[str, str]:
    if "description" not in meta or not meta["description"]:
        return assoc(warnings, "meta.description", WARNINGS["description_missing"])
    return warnings


@curry
def check_keywords(meta: Dict[str, Any], warnings: Dict[str, str]) -> Dict[str, str]:
    if "keywords" not in meta or not meta["keywords"]:
        return assoc(warnings, "meta.keywords", WARNINGS["keywords_missing"])
    return warnings


@curry
def check_links(meta: Dict[str, Any], warnings: Dict[str, str]) -> Dict[str, str]:
    if "links" not in meta or not meta["links"]:
        return assoc(warnings, "meta.links", WARNINGS["links_missing"])
    return warnings


#
# Sources
#


@curry
def check_sources(manifest: Manifest, warnings: Dict[str, str]) -> Dict[str, str]:
    if "sources" not in manifest or not manifest["sources"]:
        return assoc(warnings, "sources", WARNINGS["sources_missing"])
    return warnings


#
# Contract Types
#

# todo: validate a contract type matches source
@curry
def check_contract_types(
    manifest: Manifest, warnings: Dict[str, str]
) -> Dict[str, str]:
    if "contractTypes" not in manifest or not manifest["contractTypes"]:
        return assoc(warnings, "contractTypes", WARNINGS["contract_type_missing"])

    all_contract_type_validations = (
        (
            check_abi(contract_name, data),
            check_contract_type(contract_name, data),
            check_deployment_bytecode(contract_name, data),
            check_runtime_bytecode(contract_name, data),
            check_devdoc(contract_name, data),
            check_userdoc(contract_name, data),
        )
        for contract_name, data in manifest["contractTypes"].items()
    )
    return build(warnings, *sum(all_contract_type_validations, ()))


@curry
def check_abi(
    contract_name: str, data: Dict[str, Any], warnings: Dict[str, str]
) -> Dict[str, str]:
    if "abi" not in data or not data["abi"]:
        return assoc_in(
            warnings,
            ["contractTypes", contract_name, "abi"],
            WARNINGS["abi_missing"].format(contract_name),
        )
    return warnings


@curry
def check_contract_type(
    contract_name: str, data: Dict[str, Any], warnings: Dict[str, str]
) -> Dict[str, str]:
    if "contractType" not in data or not data["contractType"]:
        return assoc_in(
            warnings,
            ["contractTypes", contract_name, "contractType"],
            WARNINGS["contract_type_subfield_missing"].format(contract_name),
        )
    return warnings


@curry
def check_deployment_bytecode(
    contract_name: str, data: Dict[str, Any], warnings: Dict[str, str]
) -> Dict[str, str]:
    if "deploymentBytecode" not in data or not data["deploymentBytecode"]:
        return assoc_in(
            warnings,
            ["contractTypes", contract_name, "deploymentBytecode"],
            WARNINGS["deployment_bytecode_missing"].format(contract_name),
        )
    return build(
        warnings,
        check_bytecode_object(contract_name, "deployment", data["deploymentBytecode"]),
    )


@curry
def check_runtime_bytecode(
    contract_name: str, data: Dict[str, Any], warnings: Dict[str, str]
) -> Dict[str, str]:
    if "runtimeBytecode" not in data or not data["runtimeBytecode"]:
        return assoc_in(
            warnings,
            ["contractTypes", contract_name, "runtimeBytecode"],
            WARNINGS["runtime_bytecode_missing"].format(contract_name),
        )
    return build(
        warnings,
        check_bytecode_object(contract_name, "runtime", data["runtimeBytecode"]),
    )


@curry
def check_bytecode_object(
    contract_name: str,
    bytecode_type: str,
    bytecode_data: Dict[str, Any],
    warnings: Dict[str, str],
) -> Dict[str, str]:
    # todo: check if bytecode has link_refs & validate link_refs present in object
    if "bytecode" not in bytecode_data or not bytecode_data["bytecode"]:
        return assoc_in(
            warnings,
            ["contractTypes", contract_name, f"{bytecode_type}Bytecode"],
            WARNINGS["bytecode_subfield_missing"].format(contract_name, bytecode_type),
        )
    return warnings


@curry
def check_devdoc(
    contract_name: str, data: Dict[str, Any], warnings: Dict[str, str]
) -> Dict[str, str]:
    if "devdoc" not in data or not data["devdoc"]:
        return assoc_in(
            warnings,
            ["contractTypes", contract_name, "devdoc"],
            WARNINGS["devdoc_missing"].format(contract_name),
        )
    return warnings


@curry
def check_userdoc(
    contract_name: str, data: Dict[str, Any], warnings: Dict[str, str]
) -> Dict[str, str]:
    if "userdoc" not in data or not data["userdoc"]:
        return assoc_in(
            warnings,
            ["contractTypes", contract_name, "userdoc"],
            WARNINGS["userdoc_missing"].format(contract_name),
        )
    return warnings


@curry
def check_compilers(manifest: Manifest, warnings: Dict[str, str]) -> Dict[str, str]:
    if "compilers" not in manifest or not manifest["compilers"]:
        return assoc(warnings, "compilers", WARNINGS["compilers_missing"])
    return warnings



class PytestEthereumError(Exception):
    """
    Base class for all Pytest-Ethereum errors.
    """

    pass


class DeployerError(PytestEthereumError):
    """
    Raised when the Deployer is unable to deploy a contract type.
    """

    pass


class LinkerError(PytestEthereumError):
    """
    Raised when the Linker is unable to link two contract types.
    """

    pass



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
)

from eth_utils.toolz import (
    assoc,
)

from web3._utils.async_transactions import (
    get_buffered_gas_estimate as async_get_buffered_gas_estimate,
)
from web3._utils.transactions import (
    get_buffered_gas_estimate,
)
from web3.types import (
    AsyncMiddleware,
    RPCEndpoint,
    RPCResponse,
)

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401


def buffered_gas_estimate_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], w3: "Web3"
) -> Callable[[RPCEndpoint, Any], RPCResponse]:
    def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
        if method == "eth_sendTransaction":
            transaction = params[0]
            if "gas" not in transaction:
                transaction = assoc(
                    transaction,
                    "gas",
                    hex(get_buffered_gas_estimate(w3, transaction)),
                )
                return make_request(method, [transaction])
        return make_request(method, params)

    return middleware


async def async_buffered_gas_estimate_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], w3: "Web3"
) -> AsyncMiddleware:
    async def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
        if method == "eth_sendTransaction":
            transaction = params[0]
            if "gas" not in transaction:
                gas_estimate = await async_get_buffered_gas_estimate(w3, transaction)
                transaction = assoc(transaction, "gas", hex(gas_estimate))
                return await make_request(method, [transaction])
        return await make_request(method, params)

    return middleware



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
)

from eth_utils.curried import (
    apply_formatter_at_index,
    apply_formatter_if,
    apply_formatters_to_dict,
    is_null,
    is_string,
)
from eth_utils.toolz import (
    complement,
    compose,
    curry,
    dissoc,
)
from hexbytes import (
    HexBytes,
)

from web3._utils.formatters import (
    hex_to_integer,
)
from web3._utils.rpc_abi import (
    RPC,
)
from web3.exceptions import (
    ExtraDataLengthError,
    ValidationError,
)
from web3.middleware.formatting import (
    async_construct_web3_formatting_middleware,
    construct_web3_formatting_middleware,
)
from web3.types import (
    AsyncMiddleware,
    Formatters,
    FormattersDict,
    RPCEndpoint,
    TxParams,
)

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401

MAX_EXTRADATA_LENGTH = 32

is_not_null = complement(is_null)
to_integer_if_hex = apply_formatter_if(is_string, hex_to_integer)


@curry
def _validate_chain_id(web3_chain_id: int, chain_id: int) -> int:
    chain_id_int = to_integer_if_hex(chain_id)
    if chain_id_int == web3_chain_id:
        return chain_id
    else:
        raise ValidationError(
            f"The transaction declared chain ID {chain_id_int!r}, "
            f"but the connected node is on {web3_chain_id!r}"
        )


def _check_extradata_length(val: Any) -> Any:
    if not isinstance(val, (str, int, bytes)):
        return val
    result = HexBytes(val)
    if len(result) > MAX_EXTRADATA_LENGTH:
        raise ExtraDataLengthError(
            f"The field extraData is {len(result)} bytes, but should be "
            f"{MAX_EXTRADATA_LENGTH}. It is quite likely that you are "
            "connected to a POA chain. Refer to "
            "http://web3py.readthedocs.io/en/stable/middleware.html#geth-style-proof-of-authority "  # noqa: E501
            f"for more details. The full extraData is: {result!r}"
        )
    return val


def _transaction_normalizer(transaction: TxParams) -> TxParams:
    return dissoc(transaction, "chainId")


def _transaction_param_validator(web3_chain_id: int) -> Callable[..., Any]:
    transactions_params_validators = {
        "chainId": apply_formatter_if(
            # Bypass `validate_chain_id` if chainId can't be determined
            lambda _: is_not_null(web3_chain_id),
            _validate_chain_id(web3_chain_id),
        ),
    }
    return apply_formatter_at_index(
        apply_formatters_to_dict(transactions_params_validators), 0
    )


BLOCK_VALIDATORS = {
    "extraData": _check_extradata_length,
}
block_validator = apply_formatter_if(
    is_not_null, apply_formatters_to_dict(BLOCK_VALIDATORS)
)

METHODS_TO_VALIDATE = [
    RPC.eth_sendTransaction,
    RPC.eth_estimateGas,
    RPC.eth_call,
]


def _chain_id_validator(web3_chain_id: int) -> Callable[..., Any]:
    return compose(
        apply_formatter_at_index(_transaction_normalizer, 0),
        _transaction_param_validator(web3_chain_id),
    )


def _build_formatters_dict(
    request_formatters: Dict[RPCEndpoint, Any]
) -> FormattersDict:
    return dict(
        request_formatters=request_formatters,
        result_formatters={
            RPC.eth_getBlockByHash: block_validator,
            RPC.eth_getBlockByNumber: block_validator,
        },
    )


# -- sync -- #


def build_method_validators(w3: "Web3", method: RPCEndpoint) -> FormattersDict:
    request_formatters = {}
    if RPCEndpoint(method) in METHODS_TO_VALIDATE:
        w3_chain_id = w3.eth.chain_id
        for method in METHODS_TO_VALIDATE:
            request_formatters[method] = _chain_id_validator(w3_chain_id)

    return _build_formatters_dict(request_formatters)


validation_middleware = construct_web3_formatting_middleware(build_method_validators)


# -- async --- #


async def async_build_method_validators(
    async_w3: "Web3", method: RPCEndpoint
) -> FormattersDict:
    request_formatters: Formatters = {}
    if RPCEndpoint(method) in METHODS_TO_VALIDATE:
        w3_chain_id = await async_w3.eth.chain_id  # type: ignore
        for method in METHODS_TO_VALIDATE:
            request_formatters[method] = _chain_id_validator(w3_chain_id)

    return _build_formatters_dict(request_formatters)


async def async_validation_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], w3: "Web3"
) -> AsyncMiddleware:
    middleware = await async_construct_web3_formatting_middleware(
        async_build_method_validators
    )
    return await middleware(make_request, w3)



from typing import (
    Callable,
    List,
    Tuple,
)

from web3._utils.compat import (
    Protocol,
)
from web3._utils.rpc_abi import (
    RPC,
)
from web3.method import (
    DeprecatedMethod,
    Method,
    default_root_munger,
)
from web3.module import (
    Module,
)
from web3.types import (
    EnodeURI,
    NodeInfo,
    Peer,
)


def admin_start_params_munger(
    module: Module,
    host: str = "localhost",
    port: int = 8546,
    cors: str = "",
    apis: str = "eth,net,web3",
) -> Tuple[str, int, str, str]:
    return (host, port, cors, apis)


add_peer: Method[Callable[[EnodeURI], bool]] = Method(
    RPC.admin_addPeer,
    mungers=[default_root_munger],
)


datadir: Method[Callable[[], str]] = Method(
    RPC.admin_datadir,
    is_property=True,
)


node_info: Method[Callable[[], NodeInfo]] = Method(
    RPC.admin_nodeInfo,
    is_property=True,
)


peers: Method[Callable[[], List[Peer]]] = Method(
    RPC.admin_peers,
    is_property=True,
)


class ServerConnection(Protocol):
    def __call__(
        self,
        host: str = "localhost",
        port: int = 8546,
        cors: str = "",
        apis: str = "eth,net,web3",
    ) -> bool:
        pass


start_http: Method[ServerConnection] = Method(
    RPC.admin_startHTTP,
    mungers=[admin_start_params_munger],
)


start_ws: Method[ServerConnection] = Method(
    RPC.admin_startWS,
    mungers=[admin_start_params_munger],
)


stop_http: Method[Callable[[], bool]] = Method(
    RPC.admin_stopHTTP,
    is_property=True,
)


stop_ws: Method[Callable[[], bool]] = Method(
    RPC.admin_stopWS,
    is_property=True,
)

#
# Deprecated Methods
#
start_rpc = DeprecatedMethod(start_http, "start_rpc", "start_http")
stop_rpc = DeprecatedMethod(stop_http, "stop_rpc", "stop_http")



import logging
from typing import (  # noqa: F401
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)
import uuid
from uuid import UUID

from eth_utils.toolz import (
    pipe,
)
from hexbytes import (
    HexBytes,
)

from web3._utils.decorators import (
    deprecated_for,
)
from web3._utils.threads import (  # noqa: F401
    ThreadWithReturn,
    spawn,
)
from web3.datastructures import (
    NamedElementOnion,
)
from web3.exceptions import (
    BadResponseFormat,
)
from web3.middleware import (
    abi_middleware,
    attrdict_middleware,
    buffered_gas_estimate_middleware,
    gas_price_strategy_middleware,
    name_to_address_middleware,
    pythonic_middleware,
    request_parameter_normalizer,
    validation_middleware,
)
from web3.providers import (
    AutoProvider,
)
from web3.types import (  # noqa: F401
    Middleware,
    MiddlewareOnion,
    RPCEndpoint,
    RPCResponse,
)

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401
    from web3.providers import (  # noqa: F401
        AsyncBaseProvider,
        BaseProvider,
    )


NULL_RESPONSES = [None, HexBytes("0x"), "0x"]


def apply_error_formatters(
    error_formatters: Callable[..., Any],
    response: RPCResponse,
) -> RPCResponse:
    if error_formatters:
        formatted_resp = pipe(response, error_formatters)
        return formatted_resp
    else:
        return response


def apply_null_result_formatters(
    null_result_formatters: Callable[..., Any],
    response: RPCResponse,
    params: Optional[Any] = None,
) -> RPCResponse:
    if null_result_formatters:
        formatted_resp = pipe(params, null_result_formatters)
        return formatted_resp
    else:
        return response


class RequestManager:
    logger = logging.getLogger("web3.RequestManager")

    def __init__(
        self,
        w3: "Web3",
        provider: Optional[Union["BaseProvider", "AsyncBaseProvider"]] = None,
        middlewares: Optional[Sequence[Tuple[Middleware, str]]] = None,
    ) -> None:
        self.w3 = w3
        self.pending_requests: Dict[UUID, ThreadWithReturn[RPCResponse]] = {}

        if middlewares is None:
            middlewares = self.default_middlewares(w3)

        self.middleware_onion: MiddlewareOnion = NamedElementOnion(middlewares)

        if provider is None:
            self.provider = AutoProvider()
        else:
            self.provider = provider

    w3: "Web3" = None
    _provider = None

    @property
    def provider(self) -> Union["BaseProvider", "AsyncBaseProvider"]:
        return self._provider

    @provider.setter
    def provider(self, provider: Union["BaseProvider", "AsyncBaseProvider"]) -> None:
        self._provider = provider

    @staticmethod
    def default_middlewares(w3: "Web3") -> List[Tuple[Middleware, str]]:
        """
        List the default middlewares for the request manager.
        Leaving ens unspecified will prevent the middleware from resolving names.
        """
        return [
            (request_parameter_normalizer, "request_param_normalizer"),  # Delete
            (gas_price_strategy_middleware, "gas_price_strategy"),
            (name_to_address_middleware(w3), "name_to_address"),  # Add Async
            (attrdict_middleware, "attrdict"),  # Delete
            (pythonic_middleware, "pythonic"),  # Delete
            (validation_middleware, "validation"),
            (abi_middleware, "abi"),  # Delete
            (buffered_gas_estimate_middleware, "gas_estimate"),
        ]

    #
    # Provider requests and response
    #
    def _make_request(
        self, method: Union[RPCEndpoint, Callable[..., RPCEndpoint]], params: Any
    ) -> RPCResponse:
        provider = cast("BaseProvider", self.provider)
        request_func = provider.request_func(self.w3, self.middleware_onion)
        self.logger.debug(f"Making request. Method: {method}")
        return request_func(method, params)

    async def _coro_make_request(
        self, method: Union[RPCEndpoint, Callable[..., RPCEndpoint]], params: Any
    ) -> RPCResponse:
        # type ignored b/c request_func is an awaitable in async model
        request_func = await self.provider.request_func(  # type: ignore
            self.w3, self.middleware_onion
        )
        self.logger.debug(f"Making request. Method: {method}")
        return await request_func(method, params)

    @staticmethod
    def formatted_response(
        response: RPCResponse,
        params: Any,
        error_formatters: Optional[Callable[..., Any]] = None,
        null_result_formatters: Optional[Callable[..., Any]] = None,
    ) -> Any:
        if "error" in response:
            apply_error_formatters(error_formatters, response)
            raise ValueError(response["error"])
        # NULL_RESPONSES includes None, so return False here as the default
        # so we don't apply the null_result_formatters if there is no 'result' key
        elif response.get("result", False) in NULL_RESPONSES:
            # null_result_formatters raise either a BlockNotFound
            # or a TransactionNotFound error, depending on the method called
            apply_null_result_formatters(null_result_formatters, response, params)
            return response["result"]
        elif response.get("result") is not None:
            return response["result"]
        else:
            raise BadResponseFormat(
                "The response was in an unexpected format and unable to be parsed. "
                f"The raw response is: {response}"
            )

    def request_blocking(
        self,
        method: Union[RPCEndpoint, Callable[..., RPCEndpoint]],
        params: Any,
        error_formatters: Optional[Callable[..., Any]] = None,
        null_result_formatters: Optional[Callable[..., Any]] = None,
    ) -> Any:
        """
        Make a synchronous request using the provider
        """
        response = self._make_request(method, params)
        return self.formatted_response(
            response, params, error_formatters, null_result_formatters
        )

    async def coro_request(
        self,
        method: Union[RPCEndpoint, Callable[..., RPCEndpoint]],
        params: Any,
        error_formatters: Optional[Callable[..., Any]] = None,
        null_result_formatters: Optional[Callable[..., Any]] = None,
    ) -> Any:
        """
        Couroutine for making a request using the provider
        """
        response = await self._coro_make_request(method, params)
        return self.formatted_response(
            response, params, error_formatters, null_result_formatters
        )

    @deprecated_for("coro_request")
    def request_async(self, raw_method: str, raw_params: Any) -> UUID:
        request_id = uuid.uuid4()
        self.pending_requests[request_id] = spawn(
            self.request_blocking,
            raw_method=raw_method,
            raw_params=raw_params,
        )
        return request_id

    def receive_blocking(
        self, request_id: UUID, timeout: Optional[float] = None
    ) -> Any:
        try:
            request = self.pending_requests.pop(request_id)
        except KeyError:
            raise KeyError(f"Request for id:{request_id} not found")
        else:
            response = request.get(timeout=timeout)

        if "error" in response:
            raise ValueError(response["error"])

        return response["result"]

    def receive_async(self, request_id: UUID, *args: Any, **kwargs: Any) -> NoReturn:
        raise NotImplementedError("Callback pattern not implemented")



import pytest

from web3 import Web3


class GoEthereumAsyncTxPoolModuleTest:
    @pytest.mark.asyncio
    async def test_async_geth_txpool_inspect(self, async_w3: "Web3") -> None:
        test_data = await async_w3.geth.txpool.inspect()  # type: ignore
        assert "pending" in test_data

    @pytest.mark.asyncio
    async def test_async_geth_txpool_content(self, async_w3: "Web3") -> None:
        test_data = await async_w3.geth.txpool.content()  # type: ignore
        assert "pending" in test_data

    @pytest.mark.asyncio
    async def test_async_geth_txpool_status(self, async_w3: "Web3") -> None:
        test_data = await async_w3.geth.txpool.status()  # type: ignore
        assert "pending" in test_data


class GoEthereumTxPoolModuleTest:
    def test_geth_txpool_inspect(self, w3: "Web3") -> None:
        test_data = w3.geth.txpool.inspect()  # type: ignore
        assert "pending" in test_data

    def test_geth_txpool_content(self, w3: "Web3") -> None:
        test_data = w3.geth.txpool.content()  # type: ignore
        assert "pending" in test_data

    def test_geth_txpool_status(self, w3: "Web3") -> None:
        test_data = w3.geth.txpool.status()  # type: ignore
        assert "pending" in test_data



from typing import (
    Any,
    Dict,
)

from eth_abi import (
    encode_abi,
)
from eth_typing import (
    URI,
)

from web3._utils.request import (
    async_get_json_from_client_response,
    async_get_response_from_get_request,
    async_get_response_from_post_request,
)
from web3._utils.type_conversion import (
    to_bytes_if_hex,
    to_hex_if_bytes,
)
from web3.exceptions import (
    MultipleFailedRequests,
    ValidationError,
)
from web3.types import (
    TxParams,
)


async def async_handle_offchain_lookup(
    offchain_lookup_payload: Dict[str, Any],
    transaction: TxParams,
) -> bytes:
    formatted_sender = to_hex_if_bytes(offchain_lookup_payload["sender"]).lower()
    formatted_data = to_hex_if_bytes(offchain_lookup_payload["callData"]).lower()

    if formatted_sender != to_hex_if_bytes(transaction["to"]).lower():
        raise ValidationError(
            "Cannot handle OffchainLookup raised inside nested call. Returned "
            "`sender` value does not equal `to` address in transaction."
        )

    for url in offchain_lookup_payload["urls"]:
        formatted_url = URI(
            str(url)
            .replace("{sender}", str(formatted_sender))
            .replace("{data}", str(formatted_data))
        )

        try:
            if "{data}" in url and "{sender}" in url:
                response = await async_get_response_from_get_request(formatted_url)
            elif "{sender}" in url:
                response = await async_get_response_from_post_request(
                    formatted_url,
                    data={"data": formatted_data, "sender": formatted_sender},
                )
            else:
                raise ValidationError("url not formatted properly.")
        except Exception:
            continue  # try next url if timeout or issues making the request

        if (
            400 <= response.status <= 499
        ):  # if request returns 400 error, raise exception
            response.raise_for_status()
        if not 200 <= response.status <= 299:  # if not 400 error, try next url
            continue

        result = await async_get_json_from_client_response(response)

        if "data" not in result.keys():
            raise ValidationError(
                "Improperly formatted response for offchain lookup HTTP request"
                " - missing 'data' field."
            )

        encoded_data_with_function_selector = b"".join(
            [
                # 4-byte callback function selector
                to_bytes_if_hex(offchain_lookup_payload["callbackFunction"]),
                # encode the `data` from the result and the `extraData` as bytes
                encode_abi(
                    ["bytes", "bytes"],
                    [
                        to_bytes_if_hex(result["data"]),
                        to_bytes_if_hex(offchain_lookup_payload["extraData"]),
                    ],
                ),
            ]
        )

        return encoded_data_with_function_selector
    raise MultipleFailedRequests("Offchain lookup failed for supplied urls.")



# flake8: noqa

from .async_cns import (
    AsyncCNS,
)
from .base_cns import (
    BaseCNS,
)
from .cns import (
    CNS,
)

from .exceptions import (
    AddressMismatch,
    BidTooLow,
    InvalidLabel,
    InvalidName,
    UnauthorizedError,
    UnderfundedBid,
    UnownedName,
)



from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
)

from eth_typing import (
    ChecksumAddress,
    HexStr,
)
from hexbytes import (
    HexBytes,
)

from web3._utils.compat import (
    Protocol,
)
from web3._utils.rpc_abi import (
    RPC,
)
from web3.method import (
    Method,
    default_root_munger,
)
from web3.types import (
    GethWallet,
    TxParams,
)

import_raw_key: Method[Callable[[str, str], ChecksumAddress]] = Method(
    RPC.personal_importRawKey,
    mungers=[default_root_munger],
)


new_account: Method[Callable[[str], ChecksumAddress]] = Method(
    RPC.personal_newAccount,
    mungers=[default_root_munger],
)


list_accounts: Method[Callable[[], List[ChecksumAddress]]] = Method(
    RPC.personal_listAccounts,
    is_property=True,
)


list_wallets: Method[Callable[[], List[GethWallet]]] = Method(
    RPC.personal_listWallets,
    is_property=True,
)


send_transaction: Method[Callable[[TxParams, str], HexBytes]] = Method(
    RPC.personal_sendTransaction,
    mungers=[default_root_munger],
)


lock_account: Method[Callable[[ChecksumAddress], bool]] = Method(
    RPC.personal_lockAccount,
    mungers=[default_root_munger],
)


class UnlockAccountWrapper(Protocol):
    def __call__(
        self, account: ChecksumAddress, passphrase: str, duration: Optional[int] = None
    ) -> bool:
        pass


unlock_account: Method[UnlockAccountWrapper] = Method(
    RPC.personal_unlockAccount,
    mungers=[default_root_munger],
)


sign: Method[Callable[[str, ChecksumAddress, Optional[str]], HexStr]] = Method(
    RPC.personal_sign,
    mungers=[default_root_munger],
)


sign_typed_data: Method[
    Callable[[Dict[str, Any], ChecksumAddress, str], HexStr]
] = Method(
    RPC.personal_signTypedData,
    mungers=[default_root_munger],
)


ec_recover: Method[Callable[[str, HexStr], ChecksumAddress]] = Method(
    RPC.personal_ecRecover,
    mungers=[default_root_munger],
)



import operator
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
)

from eth_typing import (
    ChecksumAddress,
)
from eth_utils import (
    is_dict,
    is_hex,
    is_string,
)
from eth_utils.curried import (
    apply_formatter_if,
    apply_formatters_to_dict,
)
from eth_utils.toolz import (
    assoc,
    complement,
    compose,
    curry,
    identity,
    partial,
    pipe,
)

from web3._utils.formatters import (
    apply_formatter_to_array,
    apply_formatters_to_args,
    apply_key_map,
    hex_to_integer,
    integer_to_hex,
    is_array_of_dicts,
    static_return,
)
from web3.middleware import (
    construct_formatting_middleware,
)
from web3.middleware.formatting import (
    async_construct_formatting_middleware,
)
from web3.types import (
    Middleware,
    RPCEndpoint,
    RPCResponse,
    TxParams,
)

if TYPE_CHECKING:
    from web3 import (  # noqa: F401
        Web3,
    )


def is_named_block(value: Any) -> bool:
    return value in {"latest", "earliest", "pending"}


def is_hexstr(value: Any) -> bool:
    return is_string(value) and is_hex(value)


to_integer_if_hex = apply_formatter_if(is_hexstr, hex_to_integer)
is_not_named_block = complement(is_named_block)

# --- Request Mapping --- #

TRANSACTION_REQUEST_KEY_MAPPING = {
    "gasPrice": "gas_price",
    "maxFeePerGas": "max_fee_per_gas",
    "maxPriorityFeePerGas": "max_priority_fee_per_gas",
    "accessList": "access_list",
    "chainId": "chain_id",
}
transaction_request_remapper = apply_key_map(TRANSACTION_REQUEST_KEY_MAPPING)


TRANSACTION_REQUEST_FORMATTERS = {
    "chainId": to_integer_if_hex,
    "gas": to_integer_if_hex,
    "gasPrice": to_integer_if_hex,
    "value": to_integer_if_hex,
    "nonce": to_integer_if_hex,
    "maxFeePerGas": to_integer_if_hex,
    "maxPriorityFeePerGas": to_integer_if_hex,
    "accessList": apply_formatter_to_array(
        apply_key_map({"storageKeys": "storage_keys"})
    ),
}
transaction_request_formatter = apply_formatters_to_dict(TRANSACTION_REQUEST_FORMATTERS)

transaction_request_transformer = compose(
    transaction_request_remapper,
    transaction_request_formatter,
)

BLOCK_FORMATTERS = {
    "logsBloom": integer_to_hex,
}
block_formatter = apply_formatters_to_dict(BLOCK_FORMATTERS)


FILTER_REQUEST_KEY_MAPPING = {
    "fromBlock": "from_block",
    "toBlock": "to_block",
}
filter_request_remapper = apply_key_map(FILTER_REQUEST_KEY_MAPPING)


FILTER_REQUEST_FORMATTERS = {
    "fromBlock": to_integer_if_hex,
    "toBlock": to_integer_if_hex,
}
filter_request_formatter = apply_formatters_to_dict(FILTER_REQUEST_FORMATTERS)

filter_request_transformer = compose(
    filter_request_remapper,
    filter_request_formatter,
)


# --- Result Mapping --- #

TRANSACTION_RESULT_KEY_MAPPING = {
    "access_list": "accessList",
    "block_hash": "blockHash",
    "block_number": "blockNumber",
    "chain_id": "chainId",
    "gas_price": "gasPrice",
    "max_fee_per_gas": "maxFeePerGas",
    "max_priority_fee_per_gas": "maxPriorityFeePerGas",
    "transaction_hash": "transactionHash",
    "transaction_index": "transactionIndex",
}
transaction_result_remapper = apply_key_map(TRANSACTION_RESULT_KEY_MAPPING)


TRANSACTION_RESULT_FORMATTERS = {
    "to": apply_formatter_if(partial(operator.eq, ""), static_return(None)),
    "access_list": apply_formatter_to_array(
        apply_key_map({"storage_keys": "storageKeys"}),
    ),
}
transaction_result_formatter = apply_formatters_to_dict(TRANSACTION_RESULT_FORMATTERS)


LOG_RESULT_KEY_MAPPING = {
    "log_index": "logIndex",
    "transaction_index": "transactionIndex",
    "transaction_hash": "transactionHash",
    "block_hash": "blockHash",
    "block_number": "blockNumber",
}
log_result_remapper = apply_key_map(LOG_RESULT_KEY_MAPPING)


RECEIPT_RESULT_KEY_MAPPING = {
    "block_hash": "blockHash",
    "block_number": "blockNumber",
    "contract_address": "contractAddress",
    "gas_used": "gasUsed",
    "cumulative_gas_used": "cumulativeGasUsed",
    "effective_gas_price": "effectiveGasPrice",
    "transaction_hash": "transactionHash",
    "transaction_index": "transactionIndex",
}
receipt_result_remapper = apply_key_map(RECEIPT_RESULT_KEY_MAPPING)


BLOCK_RESULT_KEY_MAPPING = {
    "gas_limit": "gasLimit",
    "sha3_uncles": "sha3Uncles",
    "transactions_root": "transactionsRoot",
    "parent_hash": "parentHash",
    "logs_bloom": "logsBloom",
    "state_root": "stateRoot",
    "receipts_root": "receiptsRoot",
    "total_difficulty": "totalDifficulty",
    "extra_data": "extraData",
    "gas_used": "gasUsed",
    "base_fee_per_gas": "baseFeePerGas",
}
block_result_remapper = apply_key_map(BLOCK_RESULT_KEY_MAPPING)


RECEIPT_RESULT_FORMATTERS = {
    "logs": apply_formatter_to_array(log_result_remapper),
}
receipt_result_formatter = apply_formatters_to_dict(RECEIPT_RESULT_FORMATTERS)

request_formatters = {
    # Eth
    RPCEndpoint("eth_getBlockByNumber"): apply_formatters_to_args(
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
    ),
    RPCEndpoint("eth_getFilterChanges"): apply_formatters_to_args(hex_to_integer),
    RPCEndpoint("eth_getFilterLogs"): apply_formatters_to_args(hex_to_integer),
    RPCEndpoint("eth_getTransactionCount"): apply_formatters_to_args(
        identity,
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
    ),
    RPCEndpoint("eth_getBlockTransactionCountByNumber"): apply_formatters_to_args(
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
    ),
    RPCEndpoint("eth_getUncleCountByBlockNumber"): apply_formatters_to_args(
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
    ),
    RPCEndpoint("eth_getTransactionByBlockHashAndIndex"): apply_formatters_to_args(
        identity,
        to_integer_if_hex,
    ),
    RPCEndpoint("eth_getTransactionByBlockNumberAndIndex"): apply_formatters_to_args(
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
        to_integer_if_hex,
    ),
    RPCEndpoint("eth_getUncleByBlockNumberAndIndex"): apply_formatters_to_args(
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
        to_integer_if_hex,
    ),
    RPCEndpoint("eth_newFilter"): apply_formatters_to_args(
        filter_request_transformer,
    ),
    RPCEndpoint("eth_getLogs"): apply_formatters_to_args(
        filter_request_transformer,
    ),
    RPCEndpoint("eth_sendTransaction"): apply_formatters_to_args(
        transaction_request_transformer,
    ),
    RPCEndpoint("eth_estimateGas"): apply_formatters_to_args(
        transaction_request_transformer,
    ),
    RPCEndpoint("eth_call"): apply_formatters_to_args(
        transaction_request_transformer,
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
    ),
    RPCEndpoint("eth_uninstallFilter"): apply_formatters_to_args(hex_to_integer),
    RPCEndpoint("eth_getCode"): apply_formatters_to_args(
        identity,
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
    ),
    RPCEndpoint("eth_getBalance"): apply_formatters_to_args(
        identity,
        apply_formatter_if(is_not_named_block, to_integer_if_hex),
    ),
    # EVM
    RPCEndpoint("evm_revert"): apply_formatters_to_args(hex_to_integer),
    # Personal
    RPCEndpoint("personal_sendTransaction"): apply_formatters_to_args(
        transaction_request_transformer,
        identity,
    ),
}
result_formatters: Optional[Dict[RPCEndpoint, Callable[..., Any]]] = {
    RPCEndpoint("eth_getBlockByHash"): apply_formatter_if(
        is_dict,
        compose(block_result_remapper, block_formatter),
    ),
    RPCEndpoint("eth_getBlockByNumber"): apply_formatter_if(
        is_dict,
        compose(block_result_remapper, block_formatter),
    ),
    RPCEndpoint("eth_getBlockTransactionCountByHash"): apply_formatter_if(
        is_dict,
        transaction_result_remapper,
    ),
    RPCEndpoint("eth_getBlockTransactionCountByNumber"): apply_formatter_if(
        is_dict,
        transaction_result_remapper,
    ),
    RPCEndpoint("eth_getTransactionByHash"): apply_formatter_if(
        is_dict,
        compose(transaction_result_remapper, transaction_result_formatter),
    ),
    RPCEndpoint("eth_getTransactionReceipt"): apply_formatter_if(
        is_dict,
        compose(receipt_result_remapper, receipt_result_formatter),
    ),
    RPCEndpoint("eth_newFilter"): integer_to_hex,
    RPCEndpoint("eth_newBlockFilter"): integer_to_hex,
    RPCEndpoint("eth_newPendingTransactionFilter"): integer_to_hex,
    RPCEndpoint("eth_getLogs"): apply_formatter_if(
        is_array_of_dicts,
        apply_formatter_to_array(log_result_remapper),
    ),
    RPCEndpoint("eth_getFilterChanges"): apply_formatter_if(
        is_array_of_dicts,
        apply_formatter_to_array(log_result_remapper),
    ),
    RPCEndpoint("eth_getFilterLogs"): apply_formatter_if(
        is_array_of_dicts,
        apply_formatter_to_array(log_result_remapper),
    ),
    # EVM
    RPCEndpoint("evm_snapshot"): integer_to_hex,
}


async def async_ethereum_tester_middleware(  # type: ignore
    make_request, web3: "Web3"
) -> Middleware:
    middleware = await async_construct_formatting_middleware(
        request_formatters=request_formatters, result_formatters=result_formatters
    )
    return await middleware(make_request, web3)


ethereum_tester_middleware = construct_formatting_middleware(
    request_formatters=request_formatters, result_formatters=result_formatters
)


def guess_from(w3: "Web3", _: TxParams) -> ChecksumAddress:
    if w3.eth.coinbase:
        return w3.eth.coinbase
    elif w3.eth.accounts and len(w3.eth.accounts) > 0:
        return w3.eth.accounts[0]

    return None


async def async_guess_from(async_w3: "Web3", _: TxParams) -> ChecksumAddress:
    coinbase = await async_w3.eth.coinbase  # type: ignore
    accounts = await async_w3.eth.accounts  # type: ignore
    if coinbase is not None:
        return coinbase
    elif accounts is not None and len(accounts) > 0:
        return accounts[0]

    return None


@curry
def fill_default(
    field: str, guess_func: Callable[..., Any], w3: "Web3", transaction: TxParams
) -> TxParams:
    # type ignored b/c TxParams keys must be string literal types
    if field in transaction and transaction[field] is not None:  # type: ignore
        return transaction
    else:
        guess_val = guess_func(w3, transaction)
        return assoc(transaction, field, guess_val)


@curry
async def async_fill_default(
    field: str, guess_func: Callable[..., Any], async_w3: "Web3", transaction: TxParams
) -> TxParams:
    # type ignored b/c TxParams keys must be string literal types
    if field in transaction and transaction[field] is not None:  # type: ignore
        return transaction
    else:
        guess_val = await guess_func(async_w3, transaction)
        return assoc(transaction, field, guess_val)


def default_transaction_fields_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], w3: "Web3"
) -> Callable[[RPCEndpoint, Any], RPCResponse]:
    fill_default_from = fill_default("from", guess_from, w3)

    def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
        if method in (
            "eth_call",
            "eth_estimateGas",
            "eth_sendTransaction",
        ):
            filled_transaction = pipe(
                params[0],
                fill_default_from,
            )
            return make_request(method, [filled_transaction] + list(params)[1:])
        else:
            return make_request(method, params)

    return middleware


async def async_default_transaction_fields_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], async_w3: "Web3"
) -> Callable[[RPCEndpoint, Any], RPCResponse]:
    async def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
        if method in (
            "eth_call",
            "eth_estimateGas",
            "eth_sendTransaction",
        ):
            filled_transaction = await async_fill_default(
                "from", async_guess_from, async_w3, params[0]
            )
            return await make_request(method, [filled_transaction] + list(params)[1:])
        else:
            return await make_request(method, params)

    return middleware  # type: ignore



from json import (
    JSONDecodeError,
)
import logging
import os
from pathlib import (
    Path,
)
import socket
import sys
import threading
from types import (
    TracebackType,
)
from typing import (
    Any,
    Type,
    Union,
)

from web3._utils.threads import (
    Timeout,
)
from web3.types import (
    RPCEndpoint,
    RPCResponse,
)

from .base import (
    JSONBaseProvider,
)


def get_ipc_socket(ipc_path: str, timeout: float = 2.0) -> socket.socket:
    if sys.platform == "win32":
        # On Windows named pipe is used. Simulate socket with it.
        from web3._utils.windows import NamedPipe

        return NamedPipe(ipc_path)
    else:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(ipc_path)
        sock.settimeout(timeout)
        return sock


class PersistantSocket:
    sock = None

    def __init__(self, ipc_path: str) -> None:
        self.ipc_path = ipc_path

    def __enter__(self) -> socket.socket:
        if not self.ipc_path:
            raise FileNotFoundError(
                f"cannot connect to IPC socket at path: {self.ipc_path!r}"
            )

        if not self.sock:
            self.sock = self._open()
        return self.sock

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        # only close the socket if there was an error
        if exc_value is not None:
            try:
                self.sock.close()
            except Exception:
                pass
            self.sock = None

    def _open(self) -> socket.socket:
        return get_ipc_socket(self.ipc_path)

    def reset(self) -> socket.socket:
        self.sock.close()
        self.sock = self._open()
        return self.sock


# type ignored b/c missing return statement is by design here
def get_default_ipc_path() -> str:  # type: ignore
    if sys.platform == "darwin":
        ipc_path = os.path.expanduser(
            os.path.join("~", "Library", "Ethereum", "geth.ipc")
        )
        if os.path.exists(ipc_path):
            return ipc_path

        ipc_path = os.path.expanduser(
            os.path.join(
                "~",
                "Library",
                "Application Support",
                "io.parity.ethereum",
                "jsonrpc.ipc",
            )
        )
        if os.path.exists(ipc_path):
            return ipc_path

        base_trinity_path = Path("~").expanduser() / ".local" / "share" / "trinity"
        ipc_path = str(base_trinity_path / "mainnet" / "ipcs-eth1" / "jsonrpc.ipc")
        if Path(ipc_path).exists():
            return str(ipc_path)

    elif sys.platform.startswith("linux") or sys.platform.startswith("freebsd"):
        ipc_path = os.path.expanduser(os.path.join("~", ".ethereum", "geth.ipc"))
        if os.path.exists(ipc_path):
            return ipc_path

        ipc_path = os.path.expanduser(
            os.path.join("~", ".local", "share", "io.parity.ethereum", "jsonrpc.ipc")
        )
        if os.path.exists(ipc_path):
            return ipc_path

        base_trinity_path = Path("~").expanduser() / ".local" / "share" / "trinity"
        ipc_path = str(base_trinity_path / "mainnet" / "ipcs-eth1" / "jsonrpc.ipc")
        if Path(ipc_path).exists():
            return str(ipc_path)

    elif sys.platform == "win32":
        ipc_path = os.path.join("\\\\", ".", "pipe", "geth.ipc")
        if os.path.exists(ipc_path):
            return ipc_path

        ipc_path = os.path.join("\\\\", ".", "pipe", "jsonrpc.ipc")
        if os.path.exists(ipc_path):
            return ipc_path

    else:
        raise ValueError(
            f"Unsupported platform '{sys.platform}'.  Only darwin/linux/win32/"
            "freebsd are supported.  You must specify the ipc_path"
        )


# type ignored b/c missing return statement is by design here
def get_dev_ipc_path() -> str:  # type: ignore
    if os.environ.get("WEB3_PROVIDER_URI", ""):
        ipc_path = os.environ.get("WEB3_PROVIDER_URI")
        if os.path.exists(ipc_path):
            return ipc_path
    elif sys.platform == "darwin":
        tmpdir = os.environ.get("TMPDIR", "")
        ipc_path = os.path.expanduser(os.path.join(tmpdir, "geth.ipc"))
        if os.path.exists(ipc_path):
            return ipc_path

    elif sys.platform.startswith("linux") or sys.platform.startswith("freebsd"):
        ipc_path = os.path.expanduser(os.path.join("/tmp", "geth.ipc"))
        if os.path.exists(ipc_path):
            return ipc_path

    elif sys.platform == "win32":
        ipc_path = os.path.join("\\\\", ".", "pipe", "geth.ipc")
        if os.path.exists(ipc_path):
            return ipc_path

        ipc_path = os.path.join("\\\\", ".", "pipe", "jsonrpc.ipc")
        if os.path.exists(ipc_path):
            return ipc_path

    else:
        raise ValueError(
            f"Unsupported platform '{sys.platform}'.  Only darwin/linux/win32/"
            "freebsd are supported.  You must specify the ipc_path"
        )


class IPCProvider(JSONBaseProvider):
    logger = logging.getLogger("web3.providers.IPCProvider")
    _socket = None

    def __init__(
        self,
        ipc_path: Union[str, Path] = None,
        timeout: int = 10,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if ipc_path is None:
            self.ipc_path = get_default_ipc_path()
        elif isinstance(ipc_path, str) or isinstance(ipc_path, Path):
            self.ipc_path = str(Path(ipc_path).expanduser().resolve())
        else:
            raise TypeError("ipc_path must be of type string or pathlib.Path")

        self.timeout = timeout
        self._lock = threading.Lock()
        self._socket = PersistantSocket(self.ipc_path)
        super().__init__()

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.ipc_path}>"

    def make_request(self, method: RPCEndpoint, params: Any) -> RPCResponse:
        self.logger.debug(
            f"Making request IPC. Path: {self.ipc_path}, Method: {method}"
        )
        request = self.encode_rpc_request(method, params)

        with self._lock, self._socket as sock:
            try:
                sock.sendall(request)
            except BrokenPipeError:
                # one extra attempt, then give up
                sock = self._socket.reset()
                sock.sendall(request)

            raw_response = b""
            with Timeout(self.timeout) as timeout:
                while True:
                    try:
                        raw_response += sock.recv(4096)
                    except socket.timeout:
                        timeout.sleep(0)
                        continue
                    if raw_response == b"":
                        timeout.sleep(0)
                    elif has_valid_json_rpc_ending(raw_response):
                        try:
                            response = self.decode_rpc_response(raw_response)
                        except JSONDecodeError:
                            timeout.sleep(0)
                            continue
                        else:
                            return response
                    else:
                        timeout.sleep(0)
                        continue


# A valid JSON RPC response can only end in } or ] http://www.jsonrpc.org/specification
def has_valid_json_rpc_ending(raw_response: bytes) -> bool:
    stripped_raw_response = raw_response.rstrip()
    for valid_ending in [b"}", b"]"]:
        if stripped_raw_response.endswith(valid_ending):
            return True
    else:
        return False



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    NewType,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from eth_typing import (
    Address,
    BlockNumber,
    ChecksumAddress,
    Hash32,
    HexStr,
)
from hexbytes import (
    HexBytes,
)

from web3._utils.compat import (
    Literal,
    TypedDict,
)
from web3._utils.function_identifiers import (
    FallbackFn,
    ReceiveFn,
)
from web3.datastructures import (
    NamedElementOnion,
)

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401


TReturn = TypeVar("TReturn")
TParams = TypeVar("TParams")
TValue = TypeVar("TValue")

BlockParams = Literal["latest", "earliest", "pending"]
BlockIdentifier = Union[BlockParams, BlockNumber, Hash32, HexStr, HexBytes, int]
LatestBlockParam = Literal["latest"]

FunctionIdentifier = Union[str, Type[FallbackFn], Type[ReceiveFn]]

# bytes, hexbytes, or hexstr representing a 32 byte hash
_Hash32 = Union[Hash32, HexBytes, HexStr]
EnodeURI = NewType("EnodeURI", str)
ENS = NewType("ENS", str)
Nonce = NewType("Nonce", int)
RPCEndpoint = NewType("RPCEndpoint", str)
Timestamp = NewType("Timestamp", int)
Wei = NewType("Wei", int)
Formatters = Dict[RPCEndpoint, Callable[..., Any]]


class AccessListEntry(TypedDict):
    address: HexStr
    storageKeys: Sequence[HexStr]


AccessList = NewType("AccessList", Sequence[AccessListEntry])


# todo: move these to eth_typing once web3 is type hinted
class ABIEventParams(TypedDict, total=False):
    indexed: bool
    name: str
    type: str


class ABIEvent(TypedDict, total=False):
    anonymous: bool
    inputs: Sequence["ABIEventParams"]
    name: str
    type: Literal["event"]


class ABIFunctionComponents(TypedDict, total=False):
    # better typed as Sequence['ABIFunctionComponents'], but recursion isnt possible yet
    # https://github.com/python/mypy/issues/731
    components: Sequence[Any]
    name: str
    type: str


class ABIFunctionParams(TypedDict, total=False):
    components: Sequence["ABIFunctionComponents"]
    name: str
    type: str


class ABIFunction(TypedDict, total=False):
    constant: bool
    inputs: Sequence["ABIFunctionParams"]
    name: str
    outputs: Sequence["ABIFunctionParams"]
    payable: bool
    stateMutability: Literal["pure", "view", "nonpayable", "payable"]
    type: Literal["function", "constructor", "fallback", "receive"]


ABIElement = Union[ABIFunction, ABIEvent]
ABI = Sequence[Union[ABIFunction, ABIEvent]]


class EventData(TypedDict):
    address: ChecksumAddress
    args: Dict[str, Any]
    blockHash: HexBytes
    blockNumber: int
    event: str
    logIndex: int
    transactionHash: HexBytes
    transactionIndex: int


class RPCError(TypedDict):
    code: int
    message: str
    data: Optional[str]


class RPCResponse(TypedDict, total=False):
    error: Union[RPCError, str]
    id: int
    jsonrpc: Literal["2.0"]
    result: Any


Middleware = Callable[[Callable[[RPCEndpoint, Any], RPCResponse], "Web3"], Any]
AsyncMiddleware = Callable[[RPCEndpoint, Any], Coroutine[Any, Any, RPCResponse]]
MiddlewareOnion = NamedElementOnion[str, Middleware]


class FormattersDict(TypedDict, total=False):
    error_formatters: Optional[Formatters]
    request_formatters: Optional[Formatters]
    result_formatters: Optional[Formatters]


class FilterParams(TypedDict, total=False):
    address: Union[Address, ChecksumAddress, List[Address], List[ChecksumAddress]]
    blockHash: HexBytes
    fromBlock: BlockIdentifier
    toBlock: BlockIdentifier
    topics: Sequence[Optional[Union[_Hash32, Sequence[_Hash32]]]]


class FeeHistory(TypedDict):
    baseFeePerGas: List[Wei]
    gasUsedRatio: List[float]
    oldestBlock: BlockNumber
    reward: List[List[Wei]]


class LogReceipt(TypedDict):
    address: ChecksumAddress
    blockHash: HexBytes
    blockNumber: BlockNumber
    data: HexStr
    logIndex: int
    payload: HexBytes
    removed: bool
    topic: HexBytes
    topics: Sequence[HexBytes]
    transactionHash: HexBytes
    transactionIndex: int


# syntax b/c "from" keyword not allowed w/ class construction
TxData = TypedDict(
    "TxData",
    {
        "accessList": AccessList,
        "blockHash": HexBytes,
        "blockNumber": BlockNumber,
        "chainId": int,
        "data": Union[bytes, HexStr],
        "from": ChecksumAddress,
        "gas": int,
        "gasPrice": Wei,
        "maxFeePerGas": Wei,
        "maxPriorityFeePerGas": Wei,
        "hash": HexBytes,
        "input": HexStr,
        "nonce": Nonce,
        "r": HexBytes,
        "s": HexBytes,
        "to": ChecksumAddress,
        "transactionIndex": int,
        "type": Union[int, HexStr],
        "v": int,
        "value": Wei,
    },
    total=False,
)


# syntax b/c "from" keyword not allowed w/ class construction
TxParams = TypedDict(
    "TxParams",
    {
        "chainId": int,
        "data": Union[bytes, HexStr],
        # addr or ens
        "from": Union[Address, ChecksumAddress, str],
        "gas": int,
        # legacy pricing
        "gasPrice": Wei,
        # dynamic fee pricing
        "maxFeePerGas": Union[str, Wei],
        "maxPriorityFeePerGas": Union[str, Wei],
        "nonce": Nonce,
        # addr or ens
        "to": Union[Address, ChecksumAddress, str],
        "type": Union[int, HexStr],
        "value": Wei,
    },
    total=False,
)


CallOverrideParams = TypedDict(
    "CallOverrideParams",
    {
        "balance": Optional[Wei],
        "nonce": Optional[int],
        "code": Optional[Union[bytes, HexStr]],
        "state": Optional[Dict[str, Any]],
        "stateDiff": Optional[Dict[Address, Dict[str, Any]]],
    },
    total=False,
)


CallOverride = Dict[ChecksumAddress, CallOverrideParams]


GasPriceStrategy = Callable[["Web3", TxParams], Wei]


# syntax b/c "from" keyword not allowed w/ class construction
TxReceipt = TypedDict(
    "TxReceipt",
    {
        "blockHash": HexBytes,
        "blockNumber": BlockNumber,
        "contractAddress": Optional[ChecksumAddress],
        "cumulativeGasUsed": int,
        "effectiveGasPrice": Wei,
        "gasUsed": int,
        "from": ChecksumAddress,
        "logs": List[LogReceipt],
        "logsBloom": HexBytes,
        "root": HexStr,
        "status": int,
        "to": ChecksumAddress,
        "transactionHash": HexBytes,
        "transactionIndex": int,
    },
)


class SignedTx(TypedDict, total=False):
    raw: bytes
    tx: TxParams


class StorageProof(TypedDict):
    key: HexStr
    proof: Sequence[HexStr]
    value: HexBytes


class MerkleProof(TypedDict):
    address: ChecksumAddress
    accountProof: Sequence[HexStr]
    balance: int
    codeHash: HexBytes
    nonce: Nonce
    storageHash: HexBytes
    storageProof: Sequence[StorageProof]


class Protocol(TypedDict):
    difficulty: int
    head: HexStr
    network: int
    version: int


class NodeInfo(TypedDict):
    enode: EnodeURI
    id: HexStr
    ip: str
    listenAddr: str
    name: str
    ports: Dict[str, int]
    protocols: Dict[str, Protocol]


class Peer(TypedDict, total=False):
    caps: Sequence[str]
    id: HexStr
    name: str
    network: Dict[str, str]
    protocols: Dict[str, Protocol]


class SyncStatus(TypedDict):
    currentBlock: int
    highestBlock: int
    knownStates: int
    pulledStates: int
    startingBlock: int


class BlockData(TypedDict, total=False):
    baseFeePerGas: Wei
    difficulty: int
    extraData: HexBytes
    gasLimit: int
    gasUsed: int
    hash: HexBytes
    logsBloom: HexBytes
    miner: ChecksumAddress
    mixHash: HexBytes
    nonce: HexBytes
    number: BlockNumber
    parentHash: HexBytes
    receiptsRoot: HexBytes
    sha3Uncles: HexBytes
    size: int
    stateRoot: HexBytes
    timestamp: Timestamp
    totalDifficulty: int
    # list of tx hashes or of txdatas
    transactions: Union[Sequence[HexBytes], Sequence[TxData]]
    transactionsRoot: HexBytes
    uncles: Sequence[HexBytes]


class Uncle(TypedDict):
    author: ChecksumAddress
    difficulty: HexStr
    extraData: HexStr
    gasLimit: HexStr
    gasUsed: HexStr
    hash: HexBytes
    logsBloom: HexStr
    miner: HexBytes
    mixHash: HexBytes
    nonce: HexStr
    number: HexStr
    parentHash: HexBytes
    receiptsRoot: HexBytes
    sealFields: Sequence[HexStr]
    sha3Uncles: HexBytes
    size: int
    stateRoot: HexBytes
    timestamp: Timestamp
    totalDifficulty: HexStr
    transactions: Sequence[HexBytes]
    transactionsRoot: HexBytes
    uncles: Sequence[HexBytes]


#
# txpool types
#

# syntax b/c "from" keyword not allowed w/ class construction
PendingTx = TypedDict(
    "PendingTx",
    {
        "blockHash": HexBytes,
        "blockNumber": None,
        "from": ChecksumAddress,
        "gas": HexBytes,
        "maxFeePerGas": HexBytes,
        "maxPriorityFeePerGas": HexBytes,
        "gasPrice": HexBytes,
        "hash": HexBytes,
        "input": HexBytes,
        "nonce": HexBytes,
        "to": ChecksumAddress,
        "transactionIndex": None,
        "value": HexBytes,
    },
    total=False,
)


class TxPoolContent(TypedDict, total=False):
    pending: Dict[ChecksumAddress, Dict[Nonce, List[PendingTx]]]
    queued: Dict[ChecksumAddress, Dict[Nonce, List[PendingTx]]]


class TxPoolInspect(TypedDict, total=False):
    pending: Dict[ChecksumAddress, Dict[Nonce, str]]
    queued: Dict[ChecksumAddress, Dict[Nonce, str]]


class TxPoolStatus(TypedDict, total=False):
    pending: int
    queued: int


#
# web3.geth types
#


class GethWallet(TypedDict):
    accounts: Sequence[Dict[str, str]]
    status: str
    url: str


#
# web3.parity types
#

ParityBlockTrace = NewType("ParityBlockTrace", Dict[str, Any])
ParityFilterTrace = NewType("ParityFilterTrace", Dict[str, Any])
ParityMode = Literal["active", "passive", "dark", "offline"]
ParityTraceMode = Sequence[Literal["trace", "vmTrace", "stateDiff"]]


class ParityNetPeers(TypedDict):
    active: int
    connected: int
    max: int
    peers: List[Dict[Any, Any]]


class ParityFilterParams(TypedDict, total=False):
    after: int
    count: int
    fromAddress: Sequence[Union[Address, ChecksumAddress, ENS]]
    fromBlock: BlockIdentifier
    toAddress: Sequence[Union[Address, ChecksumAddress, ENS]]
    toBlock: BlockIdentifier



import functools
import threading
import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Set,
    Type,
    cast,
)

from eth_utils import (
    is_list_like,
)
import lru

from web3._utils.caching import (
    generate_cache_key,
)
from web3._utils.compat import (
    Literal,
    TypedDict,
)
from web3.types import (  # noqa: F401
    BlockData,
    BlockNumber,
    Middleware,
    RPCEndpoint,
    RPCResponse,
)

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401

SIMPLE_CACHE_RPC_WHITELIST = cast(
    Set[RPCEndpoint],
    {
        "web3_clientVersion",
        "web3_sha3",
        "net_version",
        # 'net_peerCount',
        # 'net_listening',
        "eth_protocolVersion",
        # 'eth_syncing',
        # 'eth_coinbase',
        # 'eth_mining',
        # 'eth_hashrate',
        # 'eth_gasPrice',
        # 'eth_accounts',
        # 'eth_blockNumber',
        # 'eth_getBalance',
        # 'eth_getStorageAt',
        # 'eth_getTransactionCount',
        "eth_getBlockTransactionCountByHash",
        # 'eth_getBlockTransactionCountByNumber',
        "eth_getUncleCountByBlockHash",
        # 'eth_getUncleCountByBlockNumber',
        # 'eth_getCode',
        # 'eth_sign',
        # 'eth_sendTransaction',
        # 'eth_sendRawTransaction',
        # 'eth_call',
        # 'eth_estimateGas',
        "eth_getBlockByHash",
        # 'eth_getBlockByNumber',
        "eth_getTransactionByHash",
        "eth_getTransactionByBlockHashAndIndex",
        # 'eth_getTransactionByBlockNumberAndIndex',
        # 'eth_getTransactionReceipt',
        "eth_getRawTransactionByHash",
        "eth_getUncleByBlockHashAndIndex",
        # 'eth_getUncleByBlockNumberAndIndex',
        # 'eth_getCompilers',
        # 'eth_compileLLL',
        # 'eth_compileSolidity',
        # 'eth_compileSerpent',
        # 'eth_newFilter',
        # 'eth_newBlockFilter',
        # 'eth_newPendingTransactionFilter',
        # 'eth_uninstallFilter',
        # 'eth_getFilterChanges',
        # 'eth_getFilterLogs',
        # 'eth_getLogs',
        # 'eth_getWork',
        # 'eth_submitWork',
        # 'eth_submitHashrate',
        "eth_chainId",
    },
)


def _should_cache(method: RPCEndpoint, params: Any, response: RPCResponse) -> bool:
    if "error" in response:
        return False
    elif "result" not in response:
        return False

    if response["result"] is None:
        return False
    return True


def construct_simple_cache_middleware(
    cache_class: Type[Dict[Any, Any]],
    rpc_whitelist: Collection[RPCEndpoint] = SIMPLE_CACHE_RPC_WHITELIST,
    should_cache_fn: Callable[[RPCEndpoint, Any, RPCResponse], bool] = _should_cache,
) -> Middleware:
    """
    Constructs a middleware which caches responses based on the request
    ``method`` and ``params``

    :param cache_class: Any dictionary-like object
    :param rpc_whitelist: A set of RPC methods which may have their responses cached.
    :param should_cache_fn: A callable which accepts ``method`` ``params`` and
        ``response`` and returns a boolean as to whether the response should be
        cached.
    """

    def simple_cache_middleware(
        make_request: Callable[[RPCEndpoint, Any], RPCResponse], w3: "Web3"
    ) -> Callable[[RPCEndpoint, Any], RPCResponse]:
        cache = cache_class()
        lock = threading.Lock()

        def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
            lock_acquired = (
                lock.acquire(blocking=False) if method in rpc_whitelist else False
            )

            try:
                if lock_acquired and method in rpc_whitelist:
                    cache_key = generate_cache_key((method, params))
                    if cache_key not in cache:
                        response = make_request(method, params)
                        if should_cache_fn(method, params, response):
                            cache[cache_key] = response
                        return response
                    return cache[cache_key]
                else:
                    return make_request(method, params)
            finally:
                if lock_acquired:
                    lock.release()

        return middleware

    return simple_cache_middleware


_simple_cache_middleware = construct_simple_cache_middleware(
    cache_class=cast(Type[Dict[Any, Any]], functools.partial(lru.LRU, 256)),
)


TIME_BASED_CACHE_RPC_WHITELIST = cast(
    Set[RPCEndpoint],
    {
        # 'web3_clientVersion',
        # 'web3_sha3',
        # 'net_version',
        # 'net_peerCount',
        # 'net_listening',
        # 'eth_protocolVersion',
        # 'eth_syncing',
        "eth_coinbase",
        # 'eth_mining',
        # 'eth_hashrate',
        # 'eth_gasPrice',
        "eth_accounts",
        # 'eth_blockNumber',
        # 'eth_getBalance',
        # 'eth_getStorageAt',
        # 'eth_getTransactionCount',
        # 'eth_getBlockTransactionCountByHash',
        # 'eth_getBlockTransactionCountByNumber',
        # 'eth_getUncleCountByBlockHash',
        # 'eth_getUncleCountByBlockNumber',
        # 'eth_getCode',
        # 'eth_sign',
        # 'eth_sendTransaction',
        # 'eth_sendRawTransaction',
        # 'eth_call',
        # 'eth_estimateGas',
        # 'eth_getBlockByHash',
        # 'eth_getBlockByNumber',
        # 'eth_getTransactionByHash',
        # 'eth_getTransactionByBlockHashAndIndex',
        # 'eth_getTransactionByBlockNumberAndIndex',
        # 'eth_getTransactionReceipt',
        # 'eth_getUncleByBlockHashAndIndex',
        # 'eth_getUncleByBlockNumberAndIndex',
        # 'eth_getCompilers',
        # 'eth_compileLLL',
        # 'eth_compileSolidity',
        # 'eth_compileSerpent',
        # 'eth_newFilter',
        # 'eth_newBlockFilter',
        # 'eth_newPendingTransactionFilter',
        # 'eth_uninstallFilter',
        # 'eth_getFilterChanges',
        # 'eth_getFilterLogs',
        # 'eth_getLogs',
        # 'eth_getWork',
        # 'eth_submitWork',
        # 'eth_submitHashrate',
    },
)


def construct_time_based_cache_middleware(
    cache_class: Callable[..., Dict[Any, Any]],
    cache_expire_seconds: int = 15,
    rpc_whitelist: Collection[RPCEndpoint] = TIME_BASED_CACHE_RPC_WHITELIST,
    should_cache_fn: Callable[[RPCEndpoint, Any, RPCResponse], bool] = _should_cache,
) -> Middleware:
    """
    Constructs a middleware which caches responses based on the request
    ``method`` and ``params`` for a maximum amount of time as specified

    :param cache_class: Any dictionary-like object
    :param cache_expire_seconds: The number of seconds an item may be cached
        before it should expire.
    :param rpc_whitelist: A set of RPC methods which may have their responses cached.
    :param should_cache_fn: A callable which accepts ``method`` ``params`` and
        ``response`` and returns a boolean as to whether the response should be
        cached.
    """

    def time_based_cache_middleware(
        make_request: Callable[[RPCEndpoint, Any], Any], w3: "Web3"
    ) -> Callable[[RPCEndpoint, Any], RPCResponse]:
        cache = cache_class()
        lock = threading.Lock()

        def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
            lock_acquired = (
                lock.acquire(blocking=False) if method in rpc_whitelist else False
            )

            try:
                if lock_acquired and method in rpc_whitelist:
                    cache_key = generate_cache_key((method, params))
                    if cache_key in cache:
                        # check that the cached response is not expired.
                        cached_at, cached_response = cache[cache_key]
                        cached_for = time.time() - cached_at

                        if cached_for <= cache_expire_seconds:
                            return cached_response
                        else:
                            del cache[cache_key]

                    # cache either missed or expired so make the request.
                    response = make_request(method, params)

                    if should_cache_fn(method, params, response):
                        cache[cache_key] = (time.time(), response)

                    return response
                else:
                    return make_request(method, params)
            finally:
                if lock_acquired:
                    lock.release()

        return middleware

    return time_based_cache_middleware


_time_based_cache_middleware = construct_time_based_cache_middleware(
    cache_class=functools.partial(lru.LRU, 256),
)


BLOCK_NUMBER_RPC_WHITELIST = cast(
    Set[RPCEndpoint],
    {
        # 'web3_clientVersion',
        # 'web3_sha3',
        # 'net_version',
        # 'net_peerCount',
        # 'net_listening',
        # 'eth_protocolVersion',
        # 'eth_syncing',
        # 'eth_coinbase',
        # 'eth_mining',
        # 'eth_hashrate',
        "eth_gasPrice",
        # 'eth_accounts',
        "eth_blockNumber",
        "eth_getBalance",
        "eth_getStorageAt",
        "eth_getTransactionCount",
        # 'eth_getBlockTransactionCountByHash',
        "eth_getBlockTransactionCountByNumber",
        # 'eth_getUncleCountByBlockHash',
        "eth_getUncleCountByBlockNumber",
        "eth_getCode",
        # 'eth_sign',
        # 'eth_sendTransaction',
        # 'eth_sendRawTransaction',
        "eth_call",
        "eth_estimateGas",
        # 'eth_getBlockByHash',
        "eth_getBlockByNumber",
        # 'eth_getTransactionByHash',
        # 'eth_getTransactionByBlockHashAndIndex',
        "eth_getTransactionByBlockNumberAndIndex",
        "eth_getTransactionReceipt",
        # 'eth_getUncleByBlockHashAndIndex',
        "eth_getUncleByBlockNumberAndIndex",
        # 'eth_getCompilers',
        # 'eth_compileLLL',
        # 'eth_compileSolidity',
        # 'eth_compileSerpent',
        # 'eth_newFilter',
        # 'eth_newBlockFilter',
        # 'eth_newPendingTransactionFilter',
        # 'eth_uninstallFilter',
        # 'eth_getFilterChanges',
        # 'eth_getFilterLogs',
        "eth_getLogs",
        # 'eth_getWork',
        # 'eth_submitWork',
        # 'eth_submitHashrate',
    },
)

AVG_BLOCK_TIME_KEY: Literal["avg_block_time"] = "avg_block_time"
AVG_BLOCK_SAMPLE_SIZE_KEY: Literal["avg_block_sample_size"] = "avg_block_sample_size"
AVG_BLOCK_TIME_UPDATED_AT_KEY: Literal[
    "avg_block_time_updated_at"
] = "avg_block_time_updated_at"


def _is_latest_block_number_request(method: RPCEndpoint, params: Any) -> bool:
    if method != "eth_getBlockByNumber":
        return False
    elif is_list_like(params) and tuple(params[:1]) == ("latest",):
        return True
    return False


BlockInfoCache = TypedDict(
    "BlockInfoCache",
    {
        "avg_block_time": float,
        "avg_block_sample_size": int,
        "avg_block_time_updated_at": float,
        "latest_block": BlockData,
    },
    total=False,
)


def construct_latest_block_based_cache_middleware(
    cache_class: Callable[..., Dict[Any, Any]],
    rpc_whitelist: Collection[RPCEndpoint] = BLOCK_NUMBER_RPC_WHITELIST,
    average_block_time_sample_size: int = 240,
    default_average_block_time: int = 15,
    should_cache_fn: Callable[[RPCEndpoint, Any, RPCResponse], bool] = _should_cache,
) -> Middleware:
    """
    Constructs a middleware which caches responses based on the request
    ``method``, ``params``, and the current latest block hash.

    :param cache_class: Any dictionary-like object
    :param rpc_whitelist: A set of RPC methods which may have their responses cached.
    :param average_block_time_sample_size: number of blocks to look back when computing
        average block time
    :param default_average_block_time: estimated number of seconds per block
    :param should_cache_fn: A callable which accepts ``method`` ``params`` and
        ``response`` and returns a boolean as to whether the response should be
        cached.

    .. note::
        This middleware avoids re-fetching the current latest block for each
        request by tracking the current average block time and only requesting
        a new block when the last seen latest block is older than the average
        block time.
    """

    def latest_block_based_cache_middleware(
        make_request: Callable[[RPCEndpoint, Any], Any], w3: "Web3"
    ) -> Callable[[RPCEndpoint, Any], RPCResponse]:
        cache = cache_class()
        block_info: BlockInfoCache = {}

        def _update_block_info_cache() -> None:
            avg_block_time = block_info.get(
                AVG_BLOCK_TIME_KEY, default_average_block_time
            )
            avg_block_sample_size = block_info.get(AVG_BLOCK_SAMPLE_SIZE_KEY, 0)
            avg_block_time_updated_at = block_info.get(AVG_BLOCK_TIME_UPDATED_AT_KEY, 0)

            # compute age as counted by number of blocks since the avg_block_time
            if avg_block_time == 0:
                avg_block_time_age_in_blocks: float = avg_block_sample_size
            else:
                avg_block_time_age_in_blocks = (
                    time.time() - avg_block_time_updated_at
                ) / avg_block_time

            if avg_block_time_age_in_blocks >= avg_block_sample_size:
                # If the length of time since the average block time as
                # measured by blocks is greater than or equal to the number of
                # blocks sampled then we need to recompute the average block
                # time.
                latest_block = w3.eth.get_block("latest")
                ancestor_block_number = BlockNumber(
                    max(
                        0,
                        latest_block["number"] - average_block_time_sample_size,
                    )
                )
                ancestor_block = w3.eth.get_block(ancestor_block_number)
                sample_size = latest_block["number"] - ancestor_block_number

                block_info[AVG_BLOCK_SAMPLE_SIZE_KEY] = sample_size
                if sample_size != 0:
                    block_info[AVG_BLOCK_TIME_KEY] = (
                        latest_block["timestamp"] - ancestor_block["timestamp"]
                    ) / sample_size
                else:
                    block_info[AVG_BLOCK_TIME_KEY] = avg_block_time
                block_info[AVG_BLOCK_TIME_UPDATED_AT_KEY] = time.time()

            if "latest_block" in block_info:
                latest_block = block_info["latest_block"]
                time_since_latest_block = time.time() - latest_block["timestamp"]

                # latest block is too old so update cache
                if time_since_latest_block > avg_block_time:
                    block_info["latest_block"] = w3.eth.get_block("latest")
            else:
                # latest block has not been fetched so we fetch it.
                block_info["latest_block"] = w3.eth.get_block("latest")

        lock = threading.Lock()

        def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
            lock_acquired = (
                lock.acquire(blocking=False) if method in rpc_whitelist else False
            )

            try:
                should_try_cache = (
                    lock_acquired
                    and method in rpc_whitelist
                    and not _is_latest_block_number_request(method, params)
                )
                if should_try_cache:
                    _update_block_info_cache()
                    latest_block_hash = block_info["latest_block"]["hash"]
                    cache_key = generate_cache_key((latest_block_hash, method, params))
                    if cache_key in cache:
                        return cache[cache_key]

                    response = make_request(method, params)
                    if should_cache_fn(method, params, response):
                        cache[cache_key] = response
                    return response
                else:
                    return make_request(method, params)
            finally:
                if lock_acquired:
                    lock.release()

        return middleware

    return latest_block_based_cache_middleware


_latest_block_based_cache_middleware = construct_latest_block_based_cache_middleware(
    cache_class=functools.partial(lru.LRU, 256),
    rpc_whitelist=BLOCK_NUMBER_RPC_WHITELIST,
)



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Optional,
)

from eth_utils.toolz import (
    assoc,
    merge,
)

from web3.types import (
    AsyncMiddleware,
    Formatters,
    FormattersDict,
    Literal,
    Middleware,
    RPCEndpoint,
    RPCResponse,
)

if TYPE_CHECKING:
    from web3 import Web3  # noqa: F401

FORMATTER_DEFAULTS: FormattersDict = {
    "request_formatters": {},
    "result_formatters": {},
    "error_formatters": {},
}


def _apply_response_formatters(
    method: RPCEndpoint,
    response: RPCResponse,
    result_formatters: Formatters,
    error_formatters: Formatters,
) -> RPCResponse:
    def _format_response(
        response_type: Literal["result", "error"],
        method_response_formatter: Callable[..., Any],
    ) -> RPCResponse:
        appropriate_response = response[response_type]
        return assoc(
            response, response_type, method_response_formatter(appropriate_response)
        )

    if "result" in response and method in result_formatters:
        return _format_response("result", result_formatters[method])
    elif "error" in response and method in error_formatters:
        return _format_response("error", error_formatters[method])
    else:
        return response


# --- sync -- #


def construct_formatting_middleware(
    request_formatters: Optional[Formatters] = None,
    result_formatters: Optional[Formatters] = None,
    error_formatters: Optional[Formatters] = None,
) -> Middleware:
    def ignore_web3_in_standard_formatters(
        _w3: "Web3",
        _method: RPCEndpoint,
    ) -> FormattersDict:
        return dict(
            request_formatters=request_formatters or {},
            result_formatters=result_formatters or {},
            error_formatters=error_formatters or {},
        )

    return construct_web3_formatting_middleware(ignore_web3_in_standard_formatters)


def construct_web3_formatting_middleware(
    web3_formatters_builder: Callable[["Web3", RPCEndpoint], FormattersDict],
) -> Middleware:
    def formatter_middleware(
        make_request: Callable[[RPCEndpoint, Any], Any],
        w3: "Web3",
    ) -> Callable[[RPCEndpoint, Any], RPCResponse]:
        def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
            formatters = merge(
                FORMATTER_DEFAULTS,
                web3_formatters_builder(w3, method),
            )
            request_formatters = formatters.pop("request_formatters")

            if method in request_formatters:
                formatter = request_formatters[method]
                params = formatter(params)
            response = make_request(method, params)

            return _apply_response_formatters(
                method=method, response=response, **formatters
            )

        return middleware

    return formatter_middleware


# --- async --- #


async def async_construct_formatting_middleware(
    request_formatters: Optional[Formatters] = None,
    result_formatters: Optional[Formatters] = None,
    error_formatters: Optional[Formatters] = None,
) -> Middleware:
    async def ignore_web3_in_standard_formatters(
        _w3: "Web3",
        _method: RPCEndpoint,
    ) -> FormattersDict:
        return dict(
            request_formatters=request_formatters or {},
            result_formatters=result_formatters or {},
            error_formatters=error_formatters or {},
        )

    return await async_construct_web3_formatting_middleware(
        ignore_web3_in_standard_formatters
    )


async def async_construct_web3_formatting_middleware(
    async_web3_formatters_builder: Callable[
        ["Web3", RPCEndpoint], Coroutine[Any, Any, FormattersDict]
    ]
) -> Callable[
    [Callable[[RPCEndpoint, Any], Any], "Web3"], Coroutine[Any, Any, AsyncMiddleware]
]:
    async def formatter_middleware(
        make_request: Callable[[RPCEndpoint, Any], Any],
        async_w3: "Web3",
    ) -> AsyncMiddleware:
        async def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
            formatters = merge(
                FORMATTER_DEFAULTS,
                await async_web3_formatters_builder(async_w3, method),
            )
            request_formatters = formatters.pop("request_formatters")

            if method in request_formatters:
                formatter = request_formatters[method]
                params = formatter(params)
            response = await make_request(method, params)

            return _apply_response_formatters(
                method=method, response=response, **formatters
            )

        return middleware

    return formatter_middleware



import asyncio
from collections import (
    OrderedDict,
)
from concurrent.futures import (
    ThreadPoolExecutor,
)
import contextlib
import logging
import os
import threading
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    List,
    Optional,
    Union,
)

from aiohttp import (
    ClientResponse,
    ClientSession,
    ClientTimeout,
)
from eth_typing import (
    URI,
)
import requests

from web3._utils.caching import (
    generate_cache_key,
)

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 10


class SessionCache:
    def __init__(self, size: int):
        self._size = size
        self._data: OrderedDict[str, Any] = OrderedDict()

    def cache(self, key: str, value: Any) -> Dict[str, Any]:
        evicted_items = None
        # If the key is already in the OrderedDict just update it
        # and don't evict any values. Ideally, we could still check to see
        # if there are too many items in the OrderedDict but that may rearrange
        # the order it should be unlikely that the size could grow over the limit
        if key not in self._data:
            while len(self._data) >= self._size:
                if evicted_items is None:
                    evicted_items = {}
                k, v = self._data.popitem(last=False)
                evicted_items[k] = v
        self._data[key] = value
        return evicted_items

    def get_cache_entry(self, key: str) -> Any:
        return self._data[key]

    def clear(self) -> None:
        self._data.clear()

    def __contains__(self, item: str) -> bool:
        return item in self._data

    def __len__(self) -> int:
        return len(self._data)


def get_default_http_endpoint() -> URI:
    return URI(os.environ.get("WEB3_HTTP_PROVIDER_URI", "http://localhost:8545"))


_session_cache = SessionCache(size=20)
_session_cache_lock = threading.Lock()


def cache_and_return_session(
    endpoint_uri: URI, session: requests.Session = None
) -> requests.Session:
    cache_key = generate_cache_key(endpoint_uri)

    evicted_items = None
    with _session_cache_lock:
        if cache_key not in _session_cache:
            if session is None:
                session = requests.Session()

            evicted_items = _session_cache.cache(cache_key, session)
            logger.debug(f"Session cached: {endpoint_uri}, {session}")

        cached_session = _session_cache.get_cache_entry(cache_key)

    if evicted_items is not None:
        evicted_sessions = evicted_items.values()
        for evicted_session in evicted_sessions:
            logger.debug(
                f"Session cache full. Session evicted from cache: {evicted_session}",
            )
        threading.Timer(
            DEFAULT_TIMEOUT + 0.1,
            _close_evicted_sessions,
            args=[evicted_sessions],
        ).start()

    return cached_session


def get_response_from_get_request(
    endpoint_uri: URI, *args: Any, **kwargs: Any
) -> requests.Response:
    kwargs.setdefault("timeout", DEFAULT_TIMEOUT)
    session = cache_and_return_session(endpoint_uri)
    response = session.get(endpoint_uri, *args, **kwargs)
    return response


def get_response_from_post_request(
    endpoint_uri: URI, *args: Any, **kwargs: Any
) -> requests.Response:
    kwargs.setdefault("timeout", DEFAULT_TIMEOUT)
    session = cache_and_return_session(endpoint_uri)
    response = session.post(endpoint_uri, *args, **kwargs)
    return response


def make_post_request(
    endpoint_uri: URI, data: Union[bytes, Dict[str, Any]], *args: Any, **kwargs: Any
) -> bytes:
    response = get_response_from_post_request(endpoint_uri, data=data, *args, **kwargs)
    response.raise_for_status()
    return response.content


def _close_evicted_sessions(evicted_sessions: List[requests.Session]) -> None:
    for evicted_session in evicted_sessions:
        evicted_session.close()
        logger.debug(f"Closed evicted session: {evicted_session}")


# --- async --- #


_async_session_cache = SessionCache(size=20)
_async_session_cache_lock = threading.Lock()
_pool = ThreadPoolExecutor(max_workers=1)


async def cache_and_return_async_session(
    endpoint_uri: URI,
    session: Optional[ClientSession] = None,
) -> ClientSession:
    cache_key = generate_cache_key(endpoint_uri)

    evicted_items = None
    async with async_lock(_async_session_cache_lock):
        if cache_key not in _async_session_cache:
            if session is None:
                session = ClientSession(raise_for_status=True)

            evicted_items = _async_session_cache.cache(cache_key, session)
            logger.debug(f"Async session cached: {endpoint_uri}, {session}")

        cached_session = _async_session_cache.get_cache_entry(cache_key)

    if evicted_items is not None:
        # At this point the evicted sessions are already popped out of the cache and
        # just stored in the `evicted_sessions` dict. So we can kick off a future task
        # to close them and it should be safe to pop out of the lock here.
        evicted_sessions = evicted_items.values()
        for evicted_session in evicted_sessions:
            logger.debug(
                "Async session cache full. Session evicted from cache: "
                f"{evicted_session}",
            )
        # Kick off a future task, in a separate thread, to close the evicted
        # sessions. In the case that the cache filled very quickly and some
        # sessions have been evicted before their original request has been made,
        # we set the timer to a bit more than the `DEFAULT_TIMEOUT` for a call. This
        # should make it so that any call from an evicted session can still be made
        # before the session is closed.
        threading.Timer(
            DEFAULT_TIMEOUT + 0.1,
            _async_close_evicted_sessions,
            args=[evicted_sessions],
        ).start()

    return cached_session


@contextlib.asynccontextmanager
async def async_lock(lock: threading.Lock) -> AsyncGenerator[None, None]:
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(_pool, lock.acquire)
    try:
        yield
    finally:
        lock.release()


async def async_get_response_from_get_request(
    endpoint_uri: URI, *args: Any, **kwargs: Any
) -> ClientResponse:
    kwargs.setdefault("timeout", ClientTimeout(DEFAULT_TIMEOUT))
    session = await cache_and_return_async_session(endpoint_uri)
    response = await session.get(endpoint_uri, *args, **kwargs)
    return response


async def async_get_response_from_post_request(
    endpoint_uri: URI, *args: Any, **kwargs: Any
) -> ClientResponse:
    kwargs.setdefault("timeout", ClientTimeout(DEFAULT_TIMEOUT))
    session = await cache_and_return_async_session(endpoint_uri)
    response = await session.post(endpoint_uri, *args, **kwargs)

    return response


async def async_make_post_request(
    endpoint_uri: URI, data: Union[bytes, Dict[str, Any]], *args: Any, **kwargs: Any
) -> bytes:
    response = await async_get_response_from_post_request(
        endpoint_uri, data=data, *args, **kwargs
    )
    return await response.read()


async def async_get_json_from_client_response(
    response: ClientResponse,
) -> Dict[str, Any]:
    return await response.json()


def _async_close_evicted_sessions(evicted_sessions: List[ClientSession]) -> None:
    loop = asyncio.new_event_loop()

    for evicted_session in evicted_sessions:
        loop.run_until_complete(evicted_session.close())
        logger.debug(f"Closed evicted async session: {evicted_session}")

    if any(not evicted_session.closed for evicted_session in evicted_sessions):
        logger.warning(
            f"Some evicted async sessions were not properly closed: {evicted_sessions}"
        )
    loop.close()



from typing import (
    List,
    Union,
)

from web3.types import (
    ABIEvent,
    ABIFunction,
)


def get_abi_input_names(abi: Union[ABIFunction, ABIEvent]) -> List[str]:
    if "inputs" not in abi and abi["type"] == "fallback":
        return []
    return [arg["name"] for arg in abi["inputs"]]


def get_abi_output_names(abi: Union[ABIFunction]) -> List[str]:
    if "outputs" not in abi and abi["type"] == "fallback":
        return []
    return [arg["name"] for arg in abi["outputs"]]



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Union,
)

from eth_typing import (
    URI,
)


class BaseURIBackend(ABC):
    """
    Generic backend that all URI backends are subclassed from.

    All subclasses must implement:
    can_resolve_uri, can_translate_uri, fetch_uri_contents
    """

    @abstractmethod
    def can_resolve_uri(self, uri: URI) -> bool:
        """
        Return a bool indicating whether this backend class can
        resolve the given URI to it's contents.
        """
        pass

    @abstractmethod
    def can_translate_uri(self, uri: URI) -> bool:
        """
        Return a bool indicating whether this backend class can
        translate the given URI to a corresponding content-addressed URI.
        """
        pass

    @abstractmethod
    def fetch_uri_contents(self, uri: URI) -> Union[bytes, URI]:
        """
        Fetch the contents stored at a URI.
        """
        pass



from pathlib import Path


ETHPM_DIR = Path(__file__).parent
ASSETS_DIR = ETHPM_DIR / "assets"


def get_ethpm_spec_dir() -> Path:
    ethpm_spec_dir = ETHPM_DIR / "ethpm-spec"
    v3_spec = ethpm_spec_dir / "spec" / "v3.spec.json"
    if not v3_spec.is_file():
        raise FileNotFoundError(
            "The ethpm-spec submodule is not available. "
            "Please import the submodule with `git submodule update --init`"
        )
    return ethpm_spec_dir


from .package import Package  # noqa: E402, F401
from .backends.registry import RegistryURI  # noqa: E402, F401



import ast
import operator
import random
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    NoReturn,
    Optional,
    Tuple,
    Type,
)

from eth_abi.abi import (
    decode_abi,
)
from eth_tester.exceptions import (
    BlockNotFound,
    FilterNotFound,
    TransactionFailed,
    TransactionNotFound,
    ValidationError,
)
from eth_typing import (
    HexAddress,
    HexStr,
)
from eth_utils import (
    decode_hex,
    encode_hex,
    is_null,
    keccak,
)
from eth_utils.curried import (
    apply_formatter_if,
)
from eth_utils.toolz import (
    compose,
    curry,
    excepts,
)

from web3._utils.method_formatters import (
    OFFCHAIN_LOOKUP_FIELDS,
)
from web3.exceptions import (
    OffchainLookup,
)
from web3.types import (
    LogReceipt,
    RPCResponse,
    TParams,
    TReturn,
    TValue,
    TxReceipt,
)

if TYPE_CHECKING:
    from eth_tester import (  # noqa: F401
        EthereumTester,
    )


def not_implemented(*args: Any, **kwargs: Any) -> NoReturn:
    raise NotImplementedError("RPC method not implemented")


@curry
def call_eth_tester(
    fn_name: str,
    eth_tester: "EthereumTester",
    fn_args: Any,
    fn_kwargs: Optional[Any] = None,
) -> RPCResponse:
    if fn_kwargs is None:
        fn_kwargs = {}

    try:
        return getattr(eth_tester, fn_name)(*fn_args, **fn_kwargs)
    except TransactionFailed as e:
        possible_data = e.args[0]
        if isinstance(possible_data, str) and possible_data[2:10] == "Uo\\x180\\":
            # EIP-3668 | CCIP Read
            # b"Uo\x180" is the first 4 bytes of the keccak hash for:
            # OffchainLookup(address,string[],bytes,bytes4,bytes)
            parsed_data_as_bytes = ast.literal_eval(possible_data)
            data_payload = parsed_data_as_bytes[
                4:
            ]  # everything but the function selector
            abi_decoded_data = decode_abi(OFFCHAIN_LOOKUP_FIELDS.values(), data_payload)
            offchain_lookup_payload = dict(
                zip(OFFCHAIN_LOOKUP_FIELDS.keys(), abi_decoded_data)
            )
            raise OffchainLookup(offchain_lookup_payload)
        raise e


def without_eth_tester(
    fn: Callable[[TParams], TReturn]
) -> Callable[["EthereumTester", TParams], TReturn]:
    # workaround for: https://github.com/pytoolz/cytoolz/issues/103
    # @functools.wraps(fn)
    def inner(eth_tester: "EthereumTester", params: TParams) -> TReturn:
        return fn(params)

    return inner


def without_params(
    fn: Callable[[TParams], TReturn]
) -> Callable[["EthereumTester", TParams], TReturn]:
    # workaround for: https://github.com/pytoolz/cytoolz/issues/103
    # @functools.wraps(fn)
    def inner(eth_tester: "EthereumTester", params: Any) -> TReturn:
        return fn(eth_tester)

    return inner


@curry
def preprocess_params(
    eth_tester: "EthereumTester", params: Any, preprocessor_fn: Callable[..., Any]
) -> Tuple["EthereumTester", Callable[..., Any]]:
    return eth_tester, preprocessor_fn(params)


def static_return(value: TValue) -> Callable[..., TValue]:
    def inner(*args: Any, **kwargs: Any) -> TValue:
        return value

    return inner


def client_version(eth_tester: "EthereumTester", params: Any) -> str:
    # TODO: account for the backend that is in use.
    from eth_tester import __version__

    v = sys.version_info
    return f"EthereumTester/{__version__}/{sys.platform}/python{v.major}.{v.minor}.{v.micro}"  # noqa: E501


@curry
def null_if_excepts(
    exc_type: Type[BaseException], fn: Callable[..., TReturn]
) -> Callable[..., TReturn]:
    return excepts(
        exc_type,
        fn,
        static_return(None),
    )


null_if_block_not_found = null_if_excepts(BlockNotFound)
null_if_transaction_not_found = null_if_excepts(TransactionNotFound)
null_if_filter_not_found = null_if_excepts(FilterNotFound)
null_if_indexerror = null_if_excepts(IndexError)


@null_if_indexerror
@null_if_block_not_found
def get_transaction_by_block_hash_and_index(
    eth_tester: "EthereumTester", params: Any
) -> TxReceipt:
    block_hash, transaction_index = params
    block = eth_tester.get_block_by_hash(block_hash, full_transactions=True)
    transaction = block["transactions"][transaction_index]
    return transaction


@null_if_indexerror
@null_if_block_not_found
def get_transaction_by_block_number_and_index(
    eth_tester: "EthereumTester", params: Any
) -> TxReceipt:
    block_number, transaction_index = params
    block = eth_tester.get_block_by_number(block_number, full_transactions=True)
    transaction = block["transactions"][transaction_index]
    return transaction


def create_log_filter(eth_tester: "EthereumTester", params: Any) -> int:
    filter_params = params[0]
    filter_id = eth_tester.create_log_filter(**filter_params)
    return filter_id


def get_logs(eth_tester: "EthereumTester", params: Any) -> List[LogReceipt]:
    filter_params = params[0]
    logs = eth_tester.get_logs(**filter_params)
    return logs


def _generate_random_private_key() -> HexStr:
    """
    WARNING: This is not a secure way to generate private keys and should only
    be used for testing purposes.
    """
    return encode_hex(bytes(bytearray((random.randint(0, 255) for _ in range(32)))))


@without_params
def create_new_account(eth_tester: "EthereumTester") -> HexAddress:
    return eth_tester.add_account(_generate_random_private_key())


def personal_send_transaction(eth_tester: "EthereumTester", params: Any) -> HexStr:
    transaction, password = params

    try:
        eth_tester.unlock_account(transaction["from"], password)
        transaction_hash = eth_tester.send_transaction(transaction)
    finally:
        eth_tester.lock_account(transaction["from"])

    return transaction_hash


API_ENDPOINTS = {
    "web3": {
        "clientVersion": client_version,
        "sha3": compose(
            encode_hex,
            keccak,
            decode_hex,
            without_eth_tester(operator.itemgetter(0)),
        ),
    },
    "net": {
        "version": static_return("1"),
        "listening": static_return(False),
        "peerCount": static_return(0),
    },
    "eth": {
        "protocolVersion": static_return(63),
        "syncing": static_return(False),
        "coinbase": compose(
            operator.itemgetter(0),
            call_eth_tester("get_accounts"),
        ),
        "mining": static_return(False),
        "hashrate": static_return(0),
        "chainId": static_return(131277322940537),  # from fixture generation file
        "feeHistory": not_implemented,
        "maxPriorityFeePerGas": static_return(10**9),
        "gasPrice": static_return(10**9),  # must be >= base fee post-London
        "accounts": call_eth_tester("get_accounts"),
        "blockNumber": compose(
            operator.itemgetter("number"),
            call_eth_tester(
                "get_block_by_number", fn_kwargs={"block_number": "latest"}
            ),
        ),
        "getBalance": call_eth_tester("get_balance"),
        "getStorageAt": not_implemented,
        "getProof": not_implemented,
        "getTransactionCount": call_eth_tester("get_nonce"),
        "getBlockTransactionCountByHash": null_if_block_not_found(
            compose(
                len,
                operator.itemgetter("transactions"),
                call_eth_tester("get_block_by_hash"),
            )
        ),
        "getBlockTransactionCountByNumber": null_if_block_not_found(
            compose(
                len,
                operator.itemgetter("transactions"),
                call_eth_tester("get_block_by_number"),
            )
        ),
        "getUncleCountByBlockHash": null_if_block_not_found(
            compose(
                len,
                operator.itemgetter("uncles"),
                call_eth_tester("get_block_by_hash"),
            )
        ),
        "getUncleCountByBlockNumber": null_if_block_not_found(
            compose(
                len,
                operator.itemgetter("uncles"),
                call_eth_tester("get_block_by_number"),
            )
        ),
        "getCode": call_eth_tester("get_code"),
        "sign": not_implemented,
        "signTransaction": not_implemented,
        "sendTransaction": call_eth_tester("send_transaction"),
        "sendRawTransaction": call_eth_tester("send_raw_transaction"),
        "call": call_eth_tester("call"),  # TODO: untested
        "estimateGas": call_eth_tester("estimate_gas"),  # TODO: untested
        "getBlockByHash": null_if_block_not_found(call_eth_tester("get_block_by_hash")),
        "getBlockByNumber": null_if_block_not_found(
            call_eth_tester("get_block_by_number")
        ),
        "getTransactionByHash": null_if_transaction_not_found(
            call_eth_tester("get_transaction_by_hash")
        ),
        "getTransactionByBlockHashAndIndex": get_transaction_by_block_hash_and_index,
        "getTransactionByBlockNumberAndIndex": get_transaction_by_block_number_and_index,  # noqa: E501
        "getTransactionReceipt": null_if_transaction_not_found(
            compose(
                apply_formatter_if(
                    compose(is_null, operator.itemgetter("block_number")),
                    static_return(None),
                ),
                call_eth_tester("get_transaction_receipt"),
            )
        ),
        "getUncleByBlockHashAndIndex": not_implemented,
        "getUncleByBlockNumberAndIndex": not_implemented,
        "getCompilers": not_implemented,
        "compileLLL": not_implemented,
        "compileSolidity": not_implemented,
        "compileSerpent": not_implemented,
        "newFilter": create_log_filter,
        "newBlockFilter": call_eth_tester("create_block_filter"),
        "newPendingTransactionFilter": call_eth_tester(
            "create_pending_transaction_filter"
        ),
        "uninstallFilter": excepts(
            FilterNotFound,
            compose(
                is_null,
                call_eth_tester("delete_filter"),
            ),
            static_return(False),
        ),
        "getFilterChanges": null_if_filter_not_found(
            call_eth_tester("get_only_filter_changes")
        ),
        "getFilterLogs": null_if_filter_not_found(
            call_eth_tester("get_all_filter_logs")
        ),
        "getLogs": get_logs,
        "getWork": not_implemented,
        "submitWork": not_implemented,
        "submitHashrate": not_implemented,
    },
    "db": {
        "putString": not_implemented,
        "getString": not_implemented,
        "putHex": not_implemented,
        "getHex": not_implemented,
    },
    "admin": {
        "add_peer": not_implemented,
        "datadir": not_implemented,
        "node_info": not_implemented,
        "peers": not_implemented,
        "start_http": not_implemented,
        "start_ws": not_implemented,
        "stop_http": not_implemented,
        "stop_ws": not_implemented,
        # deprecated
        "start_rpc": not_implemented,
        "stop_rpc": not_implemented,
    },
    "debug": {
        "backtraceAt": not_implemented,
        "blockProfile": not_implemented,
        "cpuProfile": not_implemented,
        "dumpBlock": not_implemented,
        "gtStats": not_implemented,
        "getBlockRLP": not_implemented,
        "goTrace": not_implemented,
        "memStats": not_implemented,
        "seedHashSign": not_implemented,
        "setBlockProfileRate": not_implemented,
        "setHead": not_implemented,
        "stacks": not_implemented,
        "startCPUProfile": not_implemented,
        "startGoTrace": not_implemented,
        "stopCPUProfile": not_implemented,
        "stopGoTrace": not_implemented,
        "traceBlock": not_implemented,
        "traceBlockByNumber": not_implemented,
        "traceBlockByHash": not_implemented,
        "traceBlockFromFile": not_implemented,
        "traceTransaction": not_implemented,
        "verbosity": not_implemented,
        "vmodule": not_implemented,
        "writeBlockProfile": not_implemented,
        "writeMemProfile": not_implemented,
    },
    "miner": {
        "make_dag": not_implemented,
        "set_extra": not_implemented,
        "set_gas_price": not_implemented,
        "start": not_implemented,
        "stop": not_implemented,
        "start_auto_dag": not_implemented,
        "stop_auto_dag": not_implemented,
        # deprecated
        "makeDAG": not_implemented,
        "setExtra": not_implemented,
        "setGasPrice": not_implemented,
        "startAutoDAG": not_implemented,
        "stopAutoDAG": not_implemented,
    },
    "personal": {
        "ec_recover": not_implemented,
        "import_raw_key": call_eth_tester("add_account"),
        "list_accounts": call_eth_tester("get_accounts"),
        "list_wallets": not_implemented,
        "lock_account": excepts(
            ValidationError,
            compose(static_return(True), call_eth_tester("lock_account")),
            static_return(False),
        ),
        "new_account": create_new_account,
        "unlock_account": excepts(
            ValidationError,
            compose(static_return(True), call_eth_tester("unlock_account")),
            static_return(False),
        ),
        "send_transaction": personal_send_transaction,
        "sign": not_implemented,
        # deprecated
        "ecRecover": not_implemented,
        "importRawKey": call_eth_tester("add_account"),
        "listAccounts": call_eth_tester("get_accounts"),
        "lockAccount": excepts(
            ValidationError,
            compose(static_return(True), call_eth_tester("lock_account")),
            static_return(False),
        ),
        "newAccount": create_new_account,
        "unlockAccount": excepts(
            ValidationError,
            compose(static_return(True), call_eth_tester("unlock_account")),
            static_return(False),
        ),
        "sendTransaction": personal_send_transaction,
    },
    "testing": {
        "timeTravel": call_eth_tester("time_travel"),
    },
    "txpool": {
        "content": not_implemented,
        "inspect": not_implemented,
        "status": not_implemented,
    },
    "evm": {
        "mine": call_eth_tester("mine_blocks"),
        "revert": call_eth_tester("revert_to_snapshot"),
        "snapshot": call_eth_tester("take_snapshot"),
    },
}



import binascii
from collections import (
    abc,
    namedtuple,
)
import copy
import itertools
import re
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)
import warnings

from eth_abi import (
    codec,
    decoding,
    encoding,
)
from eth_abi.base import (
    parse_type_str,
)
from eth_abi.exceptions import (
    ValueOutOfBounds,
)
from eth_abi.grammar import (
    ABIType,
    BasicType,
    TupleType,
    parse,
)
from eth_abi.registry import (
    ABIRegistry,
    BaseEquals,
    registry as default_registry,
)
from eth_typing import (
    HexStr,
    TypeStr,
)
from eth_utils import (
    combomethod,
    decode_hex,
    is_bytes,
    is_list_like,
    is_text,
    to_text,
    to_tuple,
)
from eth_utils.abi import (
    collapse_if_tuple,
)
from eth_utils.toolz import (
    curry,
    partial,
    pipe,
)

from web3._utils.ens import (
    is_ens_name,
)
from web3._utils.formatters import (
    recursive_map,
)
from web3.exceptions import (
    FallbackNotFound,
)
from web3.types import (
    ABI,
    ABIEvent,
    ABIEventParams,
    ABIFunction,
    ABIFunctionParams,
)
from web3.utils import (  # public utils module
    get_abi_input_names,
)


def filter_by_type(_type: str, contract_abi: ABI) -> List[Union[ABIFunction, ABIEvent]]:
    return [abi for abi in contract_abi if abi["type"] == _type]


def filter_by_name(name: str, contract_abi: ABI) -> List[Union[ABIFunction, ABIEvent]]:
    return [
        abi
        for abi in contract_abi
        if (
            abi["type"] not in ("fallback", "constructor", "receive")
            and abi["name"] == name
        )
    ]


def get_abi_input_types(abi: ABIFunction) -> List[str]:
    if "inputs" not in abi and (abi["type"] == "fallback" or abi["type"] == "receive"):
        return []
    else:
        return [collapse_if_tuple(cast(Dict[str, Any], arg)) for arg in abi["inputs"]]


def get_abi_output_types(abi: ABIFunction) -> List[str]:
    if abi["type"] == "fallback":
        return []
    else:
        return [collapse_if_tuple(cast(Dict[str, Any], arg)) for arg in abi["outputs"]]


def get_receive_func_abi(contract_abi: ABI) -> ABIFunction:
    receive_abis = filter_by_type("receive", contract_abi)
    if receive_abis:
        return cast(ABIFunction, receive_abis[0])
    else:
        raise FallbackNotFound("No receive function was found in the contract ABI.")


def get_fallback_func_abi(contract_abi: ABI) -> ABIFunction:
    fallback_abis = filter_by_type("fallback", contract_abi)
    if fallback_abis:
        return cast(ABIFunction, fallback_abis[0])
    else:
        raise FallbackNotFound("No fallback function was found in the contract ABI.")


def fallback_func_abi_exists(contract_abi: ABI) -> List[Union[ABIFunction, ABIEvent]]:
    return filter_by_type("fallback", contract_abi)


def receive_func_abi_exists(contract_abi: ABI) -> List[Union[ABIFunction, ABIEvent]]:
    return filter_by_type("receive", contract_abi)


def get_indexed_event_inputs(event_abi: ABIEvent) -> List[ABIEventParams]:
    return [arg for arg in event_abi["inputs"] if arg["indexed"] is True]


def exclude_indexed_event_inputs(event_abi: ABIEvent) -> List[ABIEventParams]:
    return [arg for arg in event_abi["inputs"] if arg["indexed"] is False]


def get_normalized_abi_arg_type(abi_arg: ABIEventParams) -> str:
    """
    Return the normalized type for the abi argument provided.
    In order to account for tuple argument types, this abstraction
    makes use of `collapse_if_tuple()` to collapse the appropriate component
    types within a tuple type, if present.
    """
    return collapse_if_tuple(dict(abi_arg))


def filter_by_argument_count(
    num_arguments: int, contract_abi: ABI
) -> List[Union[ABIFunction, ABIEvent]]:
    return [abi for abi in contract_abi if len(abi["inputs"]) == num_arguments]


def filter_by_argument_name(
    argument_names: Collection[str], contract_abi: ABI
) -> List[Union[ABIFunction, ABIEvent]]:
    return [
        abi
        for abi in contract_abi
        if set(argument_names).intersection(get_abi_input_names(abi))
        == set(argument_names)
    ]


class AddressEncoder(encoding.AddressEncoder):
    @classmethod
    def validate_value(cls, value: Any) -> None:
        if is_ens_name(value):
            return

        super().validate_value(value)


class AcceptsHexStrEncoder(encoding.BaseEncoder):
    subencoder_cls: Type[encoding.BaseEncoder] = None
    is_strict: bool = None

    def __init__(self, subencoder: encoding.BaseEncoder) -> None:
        self.subencoder = subencoder

    # type ignored b/c conflict w/ defined BaseEncoder.is_dynamic = False
    @property
    def is_dynamic(self) -> bool:  # type: ignore
        return self.subencoder.is_dynamic

    @classmethod
    def from_type_str(
        cls, abi_type: TypeStr, registry: ABIRegistry
    ) -> "AcceptsHexStrEncoder":
        subencoder_cls = cls.get_subencoder_class()
        # cast b/c expects BaseCoder but `from_type_string`
        # restricted to BaseEncoder subclasses
        subencoder = cast(
            encoding.BaseEncoder, subencoder_cls.from_type_str(abi_type, registry)
        )
        return cls(subencoder)

    @classmethod
    def get_subencoder_class(cls) -> Type[encoding.BaseEncoder]:
        if cls.subencoder_cls is None:
            raise AttributeError(f"No subencoder class is set. {cls.__name__}")
        return cls.subencoder_cls

    # type ignored b/c combomethod makes signature conflict
    # w/ defined BaseEncoder.validate_value()
    @combomethod
    def validate_value(self, value: Any) -> None:  # type: ignore
        normalized_value = self.validate_and_normalize(value)
        return self.subencoder.validate_value(normalized_value)

    def encode(self, value: Any) -> bytes:
        normalized_value = self.validate_and_normalize(value)
        return self.subencoder.encode(normalized_value)

    def validate_and_normalize(self, value: Any) -> HexStr:
        raw_value = value
        if is_text(value):
            try:
                value = decode_hex(value)
            except binascii.Error:
                self.invalidate_value(
                    value,
                    msg=f"{value} is an invalid hex string",
                )
            else:
                if raw_value[:2] != "0x":
                    if self.is_strict:
                        self.invalidate_value(
                            raw_value, msg="hex string must be prefixed with 0x"
                        )
                    elif raw_value[:2] != "0x":
                        warnings.warn(
                            "in v6 it will be invalid to pass a hex "
                            'string without the "0x" prefix',
                            category=DeprecationWarning,
                        )
        return value


class BytesEncoder(AcceptsHexStrEncoder):
    subencoder_cls = encoding.BytesEncoder
    is_strict = False


class ByteStringEncoder(AcceptsHexStrEncoder):
    subencoder_cls = encoding.ByteStringEncoder
    is_strict = False


class StrictByteStringEncoder(AcceptsHexStrEncoder):
    subencoder_cls = encoding.ByteStringEncoder
    is_strict = True


class ExactLengthBytesEncoder(encoding.BaseEncoder):
    # TODO: move this to eth-abi once the api is stabilized
    is_big_endian = False
    value_bit_size = None
    data_byte_size = None

    def validate(self) -> None:
        super().validate()

        if self.value_bit_size is None:
            raise ValueError("`value_bit_size` may not be none")
        if self.data_byte_size is None:
            raise ValueError("`data_byte_size` may not be none")
        if self.encode_fn is None:
            raise ValueError("`encode_fn` may not be none")
        if self.is_big_endian is None:
            raise ValueError("`is_big_endian` may not be none")

        if self.value_bit_size % 8 != 0:
            raise ValueError(
                f"Invalid value bit size: {self.value_bit_size}. "
                "Must be a multiple of 8"
            )

        if self.value_bit_size > self.data_byte_size * 8:
            raise ValueError("Value byte size exceeds data size")

    def encode(self, value: Any) -> bytes:
        normalized_value = self.validate_value(value)
        return self.encode_fn(normalized_value)

    # type ignored b/c conflict with defined BaseEncoder.validate_value() -> None
    def validate_value(self, value: Any) -> bytes:  # type: ignore
        if not is_bytes(value) and not is_text(value):
            self.invalidate_value(value)

        raw_value = value
        if is_text(value):
            try:
                value = decode_hex(value)
            except binascii.Error:
                self.invalidate_value(
                    value,
                    msg=f"{value} is not a valid hex string",
                )
            else:
                if raw_value[:2] != "0x":
                    self.invalidate_value(
                        raw_value, msg="hex string must be prefixed with 0x"
                    )

        byte_size = self.value_bit_size // 8
        if len(value) > byte_size:
            self.invalidate_value(
                value,
                exc=ValueOutOfBounds,
                msg=f"exceeds total byte size for bytes{byte_size} encoding",
            )
        elif len(value) < byte_size:
            self.invalidate_value(
                value,
                exc=ValueOutOfBounds,
                msg=f"less than total byte size for bytes{byte_size} encoding",
            )
        return value

    @staticmethod
    def encode_fn(value: Any) -> bytes:
        return value

    @parse_type_str("bytes")
    def from_type_str(cls, abi_type: BasicType, registry: ABIRegistry) -> bytes:
        # type ignored b/c kwargs are set in superclass init
        # Unexpected keyword argument "value_bit_size" for "__call__" of "BaseEncoder"
        return cls(  # type: ignore
            value_bit_size=abi_type.sub * 8,
            data_byte_size=abi_type.sub,
        )


class BytesDecoder(decoding.FixedByteSizeDecoder):
    # FixedByteSizeDecoder.is_big_endian is defined as None
    is_big_endian = False  # type: ignore

    # FixedByteSizeDecoder.decoder_fn is defined as None
    @staticmethod
    def decoder_fn(data: bytes) -> bytes:  # type: ignore
        return data

    @parse_type_str("bytes")
    def from_type_str(cls, abi_type: BasicType, registry: ABIRegistry) -> bytes:
        # type ignored b/c kwargs are set in superclass init
        # Unexpected keyword argument "value_bit_size" for "__call__" of "BaseDecoder"
        return cls(  # type: ignore
            value_bit_size=abi_type.sub * 8,
            data_byte_size=abi_type.sub,
        )


class TextStringEncoder(encoding.TextStringEncoder):
    @classmethod
    def validate_value(cls, value: Any) -> None:
        if is_bytes(value):
            try:
                value = to_text(value)
            except UnicodeDecodeError:
                cls.invalidate_value(
                    value,
                    msg="not decodable as unicode string",
                )

        super().validate_value(value)


def filter_by_encodability(
    abi_codec: codec.ABIEncoder,
    args: Sequence[Any],
    kwargs: Dict[str, Any],
    contract_abi: ABI,
) -> List[ABIFunction]:
    return [
        cast(ABIFunction, function_abi)
        for function_abi in contract_abi
        if check_if_arguments_can_be_encoded(
            cast(ABIFunction, function_abi), abi_codec, args, kwargs
        )
    ]


def check_if_arguments_can_be_encoded(
    function_abi: ABIFunction,
    abi_codec: codec.ABIEncoder,
    args: Sequence[Any],
    kwargs: Dict[str, Any],
) -> bool:
    try:
        arguments = merge_args_and_kwargs(function_abi, args, kwargs)
    except TypeError:
        return False

    if len(function_abi.get("inputs", [])) != len(arguments):
        return False

    try:
        types, aligned_args = get_aligned_abi_inputs(function_abi, arguments)
    except TypeError:
        return False

    return all(
        abi_codec.is_encodable(_type, arg) for _type, arg in zip(types, aligned_args)
    )


def merge_args_and_kwargs(
    function_abi: ABIFunction, args: Sequence[Any], kwargs: Dict[str, Any]
) -> Tuple[Any, ...]:
    """
    Takes a list of positional args (``args``) and a dict of keyword args
    (``kwargs``) defining values to be passed to a call to the contract function
    described by ``function_abi``.  Checks to ensure that the correct number of
    args were given, no duplicate args were given, and no unknown args were
    given.  Returns a list of argument values aligned to the order of inputs
    defined in ``function_abi``.
    """
    # Ensure the function is being applied to the correct number of args
    if len(args) + len(kwargs) != len(function_abi.get("inputs", [])):
        raise TypeError(
            f"Incorrect argument count. Expected '{len(function_abi['inputs'])}"
            f". Got '{len(args) + len(kwargs)}'"
        )

    # If no keyword args were given, we don't need to align them
    if not kwargs:
        return cast(Tuple[Any, ...], args)

    kwarg_names = set(kwargs.keys())
    sorted_arg_names = tuple(arg_abi["name"] for arg_abi in function_abi["inputs"])
    args_as_kwargs = dict(zip(sorted_arg_names, args))

    # Check for duplicate args
    duplicate_args = kwarg_names.intersection(args_as_kwargs.keys())
    if duplicate_args:
        raise TypeError(
            f"{function_abi.get('name')}() got multiple values for argument(s) "
            f"'{', '.join(duplicate_args)}'"
        )

    # Check for unknown args
    unknown_args = kwarg_names.difference(sorted_arg_names)
    if unknown_args:
        if function_abi.get("name"):
            raise TypeError(
                f"{function_abi.get('name')}() got unexpected keyword argument(s)"
                f" '{', '.join(unknown_args)}'"
            )
        raise TypeError(
            f"Type: '{function_abi.get('type')}' got unexpected keyword argument(s)"
            f" '{', '.join(unknown_args)}'"
        )

    # Sort args according to their position in the ABI and unzip them from their
    # names
    sorted_args = tuple(
        zip(
            *sorted(
                itertools.chain(kwargs.items(), args_as_kwargs.items()),
                key=lambda kv: sorted_arg_names.index(kv[0]),
            )
        )
    )

    if sorted_args:
        return sorted_args[1]
    else:
        return tuple()


TUPLE_TYPE_STR_RE = re.compile(r"^(tuple)((\[([1-9]\d*\b)?])*)??$")


def get_tuple_type_str_parts(s: str) -> Optional[Tuple[str, Optional[str]]]:
    """
    Takes a JSON ABI type string.  For tuple type strings, returns the separated
    prefix and array dimension parts.  For all other strings, returns ``None``.
    """
    match = TUPLE_TYPE_STR_RE.match(s)

    if match is not None:
        tuple_prefix = match.group(1)
        tuple_dims = match.group(2)

        return tuple_prefix, tuple_dims

    return None


def _align_abi_input(arg_abi: ABIFunctionParams, arg: Any) -> Tuple[Any, ...]:
    """
    Aligns the values of any mapping at any level of nesting in ``arg``
    according to the layout of the corresponding abi spec.
    """
    tuple_parts = get_tuple_type_str_parts(arg_abi["type"])

    if tuple_parts is None:
        # Arg is non-tuple.  Just return value.
        return arg

    tuple_prefix, tuple_dims = tuple_parts
    if tuple_dims is None:
        # Arg is non-list tuple.  Each sub arg in `arg` will be aligned
        # according to its corresponding abi.
        sub_abis = arg_abi["components"]
    else:
        # Arg is list tuple.  A non-list version of its abi will be used to
        # align each element in `arg`.
        new_abi = copy.copy(arg_abi)
        new_abi["type"] = tuple_prefix

        sub_abis = itertools.repeat(new_abi)  # type: ignore

    if isinstance(arg, abc.Mapping):
        # Arg is mapping.  Align values according to abi order.
        aligned_arg = tuple(arg[abi["name"]] for abi in sub_abis)
    else:
        aligned_arg = arg

    if not is_list_like(aligned_arg):
        raise TypeError(
            f'Expected non-string sequence for "{arg_abi.get("type")}" '
            f"component type: got {aligned_arg}"
        )

    # convert NamedTuple to regular tuple
    typing = tuple if isinstance(aligned_arg, tuple) else type(aligned_arg)

    return typing(
        _align_abi_input(sub_abi, sub_arg)
        for sub_abi, sub_arg in zip(sub_abis, aligned_arg)
    )


def get_aligned_abi_inputs(
    abi: ABIFunction, args: Union[Tuple[Any, ...], Mapping[Any, Any]]
) -> Tuple[Tuple[Any, ...], Tuple[Any, ...]]:
    """
    Takes a function ABI (``abi``) and a sequence or mapping of args (``args``).
    Returns a list of type strings for the function's inputs and a list of
    arguments which have been aligned to the layout of those types.  The args
    contained in ``args`` may contain nested mappings or sequences corresponding
    to tuple-encoded values in ``abi``.
    """
    input_abis = abi.get("inputs", [])

    if isinstance(args, abc.Mapping):
        # `args` is mapping.  Align values according to abi order.
        args = tuple(args[abi["name"]] for abi in input_abis)

    return (
        # typed dict cannot be used w/ a normal Dict
        # https://github.com/python/mypy/issues/4976
        tuple(collapse_if_tuple(abi) for abi in input_abis),  # type: ignore
        type(args)(_align_abi_input(abi, arg) for abi, arg in zip(input_abis, args)),
    )


def get_constructor_abi(contract_abi: ABI) -> ABIFunction:
    candidates = [abi for abi in contract_abi if abi["type"] == "constructor"]
    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) == 0:
        return None
    elif len(candidates) > 1:
        raise ValueError("Found multiple constructors.")
    return None


DYNAMIC_TYPES = ["bytes", "string"]

INT_SIZES = range(8, 257, 8)
BYTES_SIZES = range(1, 33)
UINT_TYPES = [f"uint{i}" for i in INT_SIZES]
INT_TYPES = [f"int{i}" for i in INT_SIZES]
BYTES_TYPES = [f"bytes{i}" for i in BYTES_SIZES] + ["bytes32.byte"]

STATIC_TYPES = list(
    itertools.chain(
        ["address", "bool"],
        UINT_TYPES,
        INT_TYPES,
        BYTES_TYPES,
    )
)

BASE_TYPE_REGEX = "|".join(
    (_type + "(?![a-z0-9])" for _type in itertools.chain(STATIC_TYPES, DYNAMIC_TYPES))
)

SUB_TYPE_REGEX = r"\[" "[0-9]*" r"\]"

TYPE_REGEX = ("^" "(?:{base_type})" "(?:(?:{sub_type})*)?" "$").format(
    base_type=BASE_TYPE_REGEX,
    sub_type=SUB_TYPE_REGEX,
)


def is_recognized_type(abi_type: TypeStr) -> bool:
    return bool(re.match(TYPE_REGEX, abi_type))


def is_bool_type(abi_type: TypeStr) -> bool:
    return abi_type == "bool"


def is_uint_type(abi_type: TypeStr) -> bool:
    return abi_type in UINT_TYPES


def is_int_type(abi_type: TypeStr) -> bool:
    return abi_type in INT_TYPES


def is_address_type(abi_type: TypeStr) -> bool:
    return abi_type == "address"


def is_bytes_type(abi_type: TypeStr) -> bool:
    return abi_type in BYTES_TYPES + ["bytes"]


def is_string_type(abi_type: TypeStr) -> bool:
    return abi_type == "string"


@curry
def is_length(target_length: int, value: abc.Sized) -> bool:
    return len(value) == target_length


def size_of_type(abi_type: TypeStr) -> int:
    """
    Returns size in bits of abi_type
    """
    if "string" in abi_type:
        return None
    if "byte" in abi_type:
        return None
    if "[" in abi_type:
        return None
    if abi_type == "bool":
        return 8
    if abi_type == "address":
        return 160
    return int(re.sub(r"\D", "", abi_type))


END_BRACKETS_OF_ARRAY_TYPE_REGEX = r"\[[^]]*\]$"


def sub_type_of_array_type(abi_type: TypeStr) -> str:
    if not is_array_type(abi_type):
        raise ValueError(f"Cannot parse subtype of nonarray abi-type: {abi_type}")

    return re.sub(END_BRACKETS_OF_ARRAY_TYPE_REGEX, "", abi_type, 1)


def length_of_array_type(abi_type: TypeStr) -> int:
    if not is_array_type(abi_type):
        raise ValueError(f"Cannot parse length of nonarray abi-type: {abi_type}")

    inner_brackets = (
        re.search(END_BRACKETS_OF_ARRAY_TYPE_REGEX, abi_type).group(0).strip("[]")
    )
    if not inner_brackets:
        return None
    else:
        return int(inner_brackets)


ARRAY_REGEX = ("^" "[a-zA-Z0-9_]+" "({sub_type})+" "$").format(sub_type=SUB_TYPE_REGEX)


def is_array_type(abi_type: TypeStr) -> bool:
    return bool(re.match(ARRAY_REGEX, abi_type))


NAME_REGEX = "[a-zA-Z_]" "[a-zA-Z0-9_]*"


ENUM_REGEX = ("^" "{lib_name}" r"\." "{enum_name}" "$").format(
    lib_name=NAME_REGEX, enum_name=NAME_REGEX
)


def is_probably_enum(abi_type: TypeStr) -> bool:
    return bool(re.match(ENUM_REGEX, abi_type))


@to_tuple
def normalize_event_input_types(
    abi_args: Collection[Union[ABIFunction, ABIEvent]]
) -> Iterable[Union[ABIFunction, ABIEvent, Dict[TypeStr, Any]]]:
    for arg in abi_args:
        if is_recognized_type(arg["type"]):
            yield arg
        elif is_probably_enum(arg["type"]):
            yield {k: "uint8" if k == "type" else v for k, v in arg.items()}
        else:
            yield arg


def abi_to_signature(abi: Union[ABIFunction, ABIEvent]) -> str:
    function_signature = "{fn_name}({fn_input_types})".format(
        fn_name=abi["name"],
        fn_input_types=",".join(
            collapse_if_tuple(dict(arg))
            for arg in normalize_event_input_types(abi.get("inputs", []))
        ),
    )
    return function_signature


########################################################
#
#  Conditionally modifying data, tagged with ABI Types
#
########################################################


@curry
def map_abi_data(
    normalizers: Sequence[Callable[[TypeStr, Any], Tuple[TypeStr, Any]]],
    types: Sequence[TypeStr],
    data: Sequence[Any],
) -> Any:
    """
    This function will apply normalizers to your data, in the
    context of the relevant types. Each normalizer is in the format:

    def normalizer(datatype, data):
        # Conditionally modify data
        return (datatype, data)

    Where datatype is a valid ABI type string, like "uint".

    In case of an array, like "bool[2]", normalizer will receive `data`
    as an iterable of typed data, like `[("bool", True), ("bool", False)]`.

    Internals
    ---

    This is accomplished by:

    1. Decorating the data tree with types
    2. Recursively mapping each of the normalizers to the data
    3. Stripping the types back out of the tree
    """
    pipeline = itertools.chain(
        [abi_data_tree(types)],
        map(data_tree_map, normalizers),
        [partial(recursive_map, strip_abi_type)],
    )

    return pipe(data, *pipeline)


@curry
def abi_data_tree(types: Sequence[TypeStr], data: Sequence[Any]) -> List[Any]:
    """
    Decorate the data tree with pairs of (type, data). The pair tuple is actually an
    ABITypedData, but can be accessed as a tuple.

    As an example:

    >>> abi_data_tree(types=["bool[2]", "uint"], data=[[True, False], 0])
    [("bool[2]", [("bool", True), ("bool", False)]), ("uint256", 0)]
    """
    return [
        abi_sub_tree(data_type, data_value)
        for data_type, data_value in zip(types, data)
    ]


@curry
def data_tree_map(
    func: Callable[[TypeStr, Any], Tuple[TypeStr, Any]], data_tree: Any
) -> "ABITypedData":
    """
    Map func to every ABITypedData element in the tree. func will
    receive two args: abi_type, and data
    """

    def map_to_typed_data(elements: Any) -> "ABITypedData":
        if isinstance(elements, ABITypedData) and elements.abi_type is not None:
            return ABITypedData(func(*elements))
        else:
            return elements

    return recursive_map(map_to_typed_data, data_tree)


class ABITypedData(namedtuple("ABITypedData", "abi_type, data")):
    """
    This class marks data as having a certain ABI-type.

    >>> a1 = ABITypedData(['address', addr1])
    >>> a2 = ABITypedData(['address', addr2])
    >>> addrs = ABITypedData(['address[]', [a1, a2]])

    You can access the fields using tuple() interface, or with
    attributes:

    >>> assert a1.abi_type == a1[0]
    >>> assert a1.data == a1[1]

    Unlike a typical `namedtuple`, you initialize with a single
    positional argument that is iterable, to match the init
    interface of all other relevant collections.
    """

    def __new__(cls, iterable: Iterable[Any]) -> "ABITypedData":
        return super().__new__(cls, *iterable)


def abi_sub_tree(
    type_str_or_abi_type: Optional[Union[TypeStr, ABIType]], data_value: Any
) -> ABITypedData:
    if type_str_or_abi_type is None:
        return ABITypedData([None, data_value])

    if isinstance(type_str_or_abi_type, TypeStr):
        abi_type = parse(type_str_or_abi_type)
    else:
        abi_type = type_str_or_abi_type

    # In the two special cases below, we rebuild the given data structures with
    # annotated items
    if abi_type.is_array:
        # If type is array, determine item type and annotate all
        # items in iterable with that type
        item_type_str = abi_type.item_type.to_type_str()
        value_to_annotate = [
            abi_sub_tree(item_type_str, item_value) for item_value in data_value
        ]
    elif isinstance(abi_type, TupleType):
        # Otherwise, if type is tuple, determine component types and annotate
        # tuple components in iterable respectively with those types
        value_to_annotate = type(data_value)(
            abi_sub_tree(comp_type.to_type_str(), comp_value)
            for comp_type, comp_value in zip(abi_type.components, data_value)
        )
    else:
        value_to_annotate = data_value

    return ABITypedData(
        [
            abi_type.to_type_str(),
            value_to_annotate,
        ]
    )


def strip_abi_type(elements: Any) -> Any:
    if isinstance(elements, ABITypedData):
        return elements.data
    else:
        return elements


def build_default_registry() -> ABIRegistry:
    # We make a copy here just to make sure that eth-abi's default registry is not
    # affected by our custom encoder subclasses
    registry = default_registry.copy()

    registry.unregister("address")
    registry.unregister("bytes<M>")
    registry.unregister("bytes")
    registry.unregister("string")

    registry.register(
        BaseEquals("address"),
        AddressEncoder,
        decoding.AddressDecoder,
        label="address",
    )
    registry.register(
        BaseEquals("bytes", with_sub=True),
        BytesEncoder,
        decoding.BytesDecoder,
        label="bytes<M>",
    )
    registry.register(
        BaseEquals("bytes", with_sub=False),
        ByteStringEncoder,
        decoding.ByteStringDecoder,
        label="bytes",
    )
    registry.register(
        BaseEquals("string"),
        TextStringEncoder,
        decoding.StringDecoder,
        label="string",
    )
    return registry


def build_strict_registry() -> ABIRegistry:
    registry = default_registry.copy()

    registry.unregister("address")
    registry.unregister("bytes<M>")
    registry.unregister("bytes")
    registry.unregister("string")

    registry.register(
        BaseEquals("address"),
        AddressEncoder,
        decoding.AddressDecoder,
        label="address",
    )
    registry.register(
        BaseEquals("bytes", with_sub=True),
        ExactLengthBytesEncoder,
        BytesDecoder,
        label="bytes<M>",
    )
    registry.register(
        BaseEquals("bytes", with_sub=False),
        StrictByteStringEncoder,
        decoding.ByteStringDecoder,
        label="bytes",
    )
    registry.register(
        BaseEquals("string"),
        TextStringEncoder,
        decoding.StringDecoder,
        label="string",
    )
    return registry
